import tkinter as tk
from tkinter import ttk, messagebox as msg, filedialog, scrolledtext
import typing
import os
import threading
import webbrowser
import folium
import sys

from ...data_processing.era5_download import download_and_process_era_data
from ...data_processing.noaa_download import *
from ..tooltip import CreateToolTip

if typing.TYPE_CHECKING:
    from ..start_page import StartPage

class StationChoicePopup(tk.Toplevel):
    """
    Uma janela pop-up modal para o usuário selecionar uma estação de uma lista.
    """
    def __init__(self, parent, controller, stations):
        super().__init__(parent)
        self.controller = controller
        self.stations = stations
        self.chosen_station = None

        self.title(controller.i18n.get('choose_station_title'))
        self.geometry("600x400")
        
        # bloqueia a janela principal
        self.grab_set()
        self.transient(parent)

        i18n = self.controller.i18n

        # UI
        prompt_label = ttk.Label(self, text=controller.i18n.get('choose_station_prompt'), wraplength=580)
        prompt_label.pack(pady=10, padx=10)

        tree_frame = ttk.Frame(self)
        tree_frame.pack(fill=tk.BOTH, expand=True, padx=10, pady=(0, 10))
        
        scrollbar = ttk.Scrollbar(tree_frame, orient=tk.VERTICAL)
        self.tree = ttk.Treeview(
            tree_frame,
            columns=("name", "id", "enddate"),
            show="headings",
            yscrollcommand=scrollbar.set
        )
        scrollbar.config(command=self.tree.yview)

        # Cabeçalhos
        i18n_name = controller.i18n.get('station_name_header')
        i18n_id = controller.i18n.get('station_id_header')
        i18n_date = controller.i18n.get('station_end_date_header')
        
        self.tree.heading("name", text=i18n_name)
        self.tree.heading("id", text=i18n_id)
        self.tree.heading("enddate", text=i18n_date)
        self.tree.column("name", width=300, anchor="w")
        self.tree.column("id", width=150, anchor="w")
        self.tree.column("enddate", width=100, anchor="center")

        for i, station in enumerate(self.stations):
            self.tree.insert("", "end", iid=i, values=(station['name'], station['id'], station['maxdate']))

        scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        self.tree.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

        self.tree.bind("<Double-1>", self.on_choice)

    def on_choice(self, event=None):
        """Chamado quando o usuário clica duas vezes em um item."""
        try:
            selected_iid = self.tree.selection()[0]
            self.chosen_station = self.stations[int(selected_iid)]
            self.destroy()
        except (IndexError, ValueError):
            pass

class DownloadDataPage(ttk.Frame):
    def __init__(self, parent, controller):
        ttk.Frame.__init__(self, parent)
        self.controller = controller
        i18n = self.controller.i18n

        top_frame = ttk.Frame(self)
        top_frame.pack(fill=tk.X, padx=10, pady=10)

        self.back_button = ttk.Button(top_frame, text="", command=self.go_to_start_page)
        self.back_button.pack(side=tk.LEFT)

        self.page_title = ttk.Label(top_frame, text="", font=("Verdana", 16, "bold"))
        self.page_title.pack(side=tk.LEFT, expand=True)

        main_container = ttk.Frame(self)
        main_container.pack(fill=tk.BOTH, expand=True, padx=10, pady=10)
        main_container.grid_columnconfigure(0, weight=1, minsize=280) # Painel esquerdo
        main_container.grid_columnconfigure(1, weight=3) # Painel direito
        main_container.grid_rowconfigure(0, weight=1)
        
        # --- PAINEL ESQUERDO ---
        left_panel = ttk.Frame(main_container)
        left_panel.grid(row=0, column=0, sticky="nsew", padx=(0, 10))
        
        self.method_label = ttk.Label(left_panel, text="")
        self.method_label.pack(anchor="w", padx=5)
        self.method_combo = ttk.Combobox(left_panel, values=["NOAA", "ECMWF"], state="readonly")
        self.method_combo.pack(fill=tk.X, padx=5, pady=(0, 10))
        self.method_hint = CreateToolTip(self.method_combo, text=i18n.get('download_method_hint'))
        self.method_combo.bind("<<ComboboxSelected>>", self._on_method_selected)

        self.city_label = ttk.Label(left_panel, text="")
        self.city_label.pack(anchor="w", padx=5)
        self.city_entry = ttk.Entry(left_panel)
        self.city_entry.pack(fill=tk.X, padx=5, pady=(0, 10))
        self._add_placeholder(self.city_entry, "City, Country")
        self.city_hint = CreateToolTip(self.city_entry, text=i18n.get('city_download_entry_hint'))

        self.period_frame = ttk.LabelFrame(left_panel, text="")
        self.period_frame.pack(fill=tk.X, padx=5, pady=(0, 10))
        self.period_frame.grid_columnconfigure(0, weight=1)
        self.period_frame.grid_columnconfigure(1, weight=1)

        self.start_label = ttk.Label(self.period_frame, text="")
        self.start_label.grid(row=0, column=0, sticky="w", padx=5)
        self.start_entry = ttk.Entry(self.period_frame)
        self.start_entry.grid(row=1, column=0, sticky="ew", padx=5, pady=(0, 5))
        self._add_placeholder(self.start_entry, "YYYY-MM-DD")
        self.start_hint = CreateToolTip(self.start_entry, text=i18n.get('start_date_download_hint'))

        self.end_label = ttk.Label(self.period_frame, text="")
        self.end_label.grid(row=0, column=1, sticky="w", padx=5)
        self.end_entry = ttk.Entry(self.period_frame)
        self.end_entry.grid(row=1, column=1, sticky="ew", padx=5, pady=(0, 5))
        self._add_placeholder(self.end_entry, "YYYY-MM-DD")
        self.end_hint = CreateToolTip(self.end_entry, text=i18n.get('end_date_download_hint'))

        # --- Frame Condicional para o Radius ---
        self.radius_frame = ttk.Frame(left_panel)

        self.radius_label = ttk.Label(self.radius_frame, text="")
        self.radius_label.pack(anchor="w", padx=5)
        self.radius_entry = ttk.Entry(self.radius_frame)
        self.radius_entry.pack(fill=tk.X, padx=5, pady=(0, 10))
        self.radius_hint = CreateToolTip(self.radius_entry, text=i18n.get('radius_download_hint'))

        self.download_btn = ttk.Button(left_panel, text="", command=self._on_start_download)
        self.download_btn.pack(fill=tk.X, padx=5, pady=20)

        # --- PAINEL DIREITO ---
        self.right_panel = ttk.Frame(main_container, relief="solid", borderwidth=1)
        self.right_panel.grid(row=0, column=1, sticky="nsew")
        
        self.log_label = ttk.Label(self.right_panel, text="", font=("Verdana", 10, "bold"))
        self.log_label.pack(anchor="w", padx=10, pady=(0, 5))
        
        self.log_text = scrolledtext.ScrolledText(self.right_panel, state="disabled", height=10)
        self.log_text.pack(fill=tk.BOTH, expand=True, padx=10, pady=(0, 10))

        self.update_texts()
        self._on_method_selected(None)

    def _add_placeholder(self, entry, placeholder_text, color='grey'):
        default_fg = 'black'

        def on_focus_in(event):
            if entry.get() == placeholder_text and str(entry.cget('foreground')) == color:
                entry.delete(0, tk.END)
                entry.config(foreground=default_fg)

        def on_focus_out(event):
            if not entry.get():
                entry.insert(0, placeholder_text)
                entry.config(foreground=color)

        entry.bind("<FocusIn>", on_focus_in)
        entry.bind("<FocusOut>", on_focus_out)

        entry.insert(0, placeholder_text)
        entry.config(foreground=color)

    def _on_method_selected(self, event=None):
        """Chamado quando a combobox de método é alterada."""
        selected_method = self.method_combo.get()
        
        if selected_method == "NOAA":
            self.radius_frame.pack(fill=tk.X, padx=5, pady=(0, 10), before=self.download_btn)
        else:
            self.radius_frame.pack_forget()

    def _log(self, message):
        """Adiciona uma mensagem à caixa de log na UI, de forma segura entre threads."""
        self.log_text.config(state="normal")
        self.log_text.insert(tk.END, message + "\n")
        self.log_text.config(state="disabled")
        self.log_text.see(tk.END)

    def _on_start_download(self):
        """Função chamada pelo botão 'Iniciar download'."""
        #   Coleta os dados e chama sua lógica de download
        method = self.method_combo.get()
        raw_city = self.city_entry.get()
        start = self.start_entry.get()
        end = self.end_entry.get()

        if raw_city == "City, Country":
            city = ""
        else:
            city = raw_city

        if start == "YYYY-MM-DD":
            start = ""
        if end == "YYYY-MM-DD":
            end = ""

        i18n = self.controller.i18n
        
        if not all([method, city, start, end]):
             msg.showwarning(
                 title=i18n.get('missing_fields_title'),
                 message=i18n.get('all_fields_required_msg')
             )
             return

        radius = None
        if method == "NOAA":
            radius_str = self.radius_entry.get()
            if not radius_str:
                self.controller.show_translated_message('warning', 'missing_fields_title', 'radius_required_msg')
                return
            try:
                radius = float(radius_str)
            except ValueError:
                 self.controller.show_translated_message('error', 'value_error_title', 'invalid_radius_msg') # Crie esta chave JSON
                 return
        
        # Limpa o log e desabilita o botão
        self.log_text.config(state="normal")
        self.log_text.delete(1.0, tk.END)
        self.log_text.config(state="disabled")
        self.download_btn.config(state="disabled", text=i18n.get('download_in_progress'))
        
        threading.Thread(
            target=self._download_thread_target,
            args=(method, city, start, end, radius),
            daemon=True
        ).start()

    def _run_noaa_download(self, city, start_date, end_date, radius):
        """Lógica específica do NOAA, agora adaptada para a UI."""
        i18n = self.controller.i18n
        required_types = ["TMIN", "TMAX", "PRCP"]
        
        self.after(0, self._log, f"Buscando coordenadas para '{city}'...")
        lat, lon = get_city_coords(city)
        if lat is None or lon is None:
            self.after(0, self.controller.show_translated_message, 'error', 'coords_error_title', 'coords_error_msg', city=city)
            raise Exception("Coordenadas não encontradas.")
        
        self.after(0, self._log, f"Coordenadas encontradas: {lat:.4f}, {lon:.4f}")
        self.after(0, self._log, f"Buscando estações (Raio: {radius})...")

        candidate_stations = find_stations(lat, lon, 50, required_types, radius)
        if not candidate_stations:
            self.after(0, self.controller.show_translated_message, 'warning', 'no_stations_found_title', 'no_stations_found_msg')
            raise Exception("Nenhuma estação candidata encontrada.")

        valid_stations = []
        self.after(0, self._log, "Verificando dados das estações...")
        for station in candidate_stations:
            if check_station_data_types(station['id'], required_types, start_date, end_date):
                valid_stations.append(station)
                self.after(0, self._log, f"  ✔️ Estação válida: {station['name']} ({station['id']})")
            else:
                self.after(0, self._log, f"  ❌ Estação incompleta: {station['name']} ({station['id']})")

        if not valid_stations:
            self.after(0, self.controller.show_translated_message, 'warning', 'no_stations_found_title', 'no_stations_found_msg')
            raise Exception("Nenhuma estação válida encontrada.")

        # Abre o mapa e a janela pop-up
        map_file = self._create_map(valid_stations)
        webbrowser.open('file://' + os.path.realpath(map_file))
        
        popup = StationChoicePopup(self, self.controller, valid_stations)
        self.wait_window(popup) # Pausa esta thread
        
        chosen_station = popup.chosen_station

        if not chosen_station:
            self.after(0, self.controller.show_translated_message, 'info', 'no_station_chosen_title', 'no_station_chosen') # Crie 'no_station_chosen_title'
            raise Exception("Nenhuma estação escolhida.")

        self.after(0, self._log, f"Estação escolhida: '{chosen_station['name']}' ({chosen_station['id']})")
        self.after(0, self._log, "Baixando dados climáticos...")
        
        station_info = {
            'nome': chosen_station['name'], 'id': chosen_station['id'],
            'latitude': chosen_station.get('latitude', lat), 'longitude': chosen_station.get('longitude', lon),
            'elevation': chosen_station.get('elevation', 'N/A'),
            'data_inicial': start_date, 'data_final': end_date
        }
        
        data = get_dly_climate_data_for_station(chosen_station['id'], start_date, end_date, verbose=False) # verbose=False para não poluir log

        if data:
            temp_name = city.split(",")[0].replace(" ", "")
            file_name = f"{temp_name}_{chosen_station['id'].replace(':', '_')}.csv"
            
            # Salva na pasta de dados da aplicação
            base_dir = os.path.abspath(os.path.dirname(sys.argv[0]))
            output_path = os.path.join(base_dir, file_name) # Salva na raiz do projeto
            
            salvar_dados_climaticos_csv(data, output_path, station_info)
            self.after(0, self._log, f"Dados salvos com sucesso em: {output_path}")
        else:
            raise Exception(f"Erro ao baixar os dados da estação: {chosen_station['id']}")

    def _download_thread_target(self, method, city, start, end, radius):
        """A função que roda na thread separada para não congelar a UI."""
        i18n = self.controller.i18n

        base_dir = os.path.abspath(os.path.dirname(sys.argv[0]))
        output_dir = os.path.join(base_dir, "downloaded_data") # Salva tudo numa pasta 'downloaded_data'
        os.makedirs(output_dir, exist_ok=True)

        def gui_logger(message):
            self.after(0, self._log, message)

        try:
            if method == "NOAA":
                self._run_noaa_download(city, start, end, radius)
            else:
                gui_logger("Iniciando download do ECMWF...")

                download_and_process_era_data(
                    city, 
                    start, 
                    end, 
                    output_dir,  # Passa o diretório de saída
                    logger=gui_logger # Passa a nossa função de log
                )
            
            self.after(0, self.download_btn.config, {"state": "normal", "text": i18n.get('start_download_btn')})
            self.after(0, self.controller.show_translated_message, 'info', 'success_title', 'download_complete')

        except Exception as e:
            # Mostra o erro no log e reabilita o botão
            self.after(0, self._log, f"ERRO: {e}")
            self.after(0, self.download_btn.config, {"state": "normal", "text": i18n.get('start_download_btn')})
            self.after(0, self.controller.show_translated_message, 'error', 'download_failed_title', 'download_failed') # Crie 'download_failed_title'

    def _create_map(self, stations):
        """Cria e salva o mapa Folium. Retorna o caminho do arquivo."""
        map_center = [stations[0]['latitude'], stations[0]['longitude']]
        m = folium.Map(location=map_center, zoom_start=10)
        
        for i, station in enumerate(stations):
            station_id_display = i + 1 # Usa o índice + 1 para o ID de escolha
            popup_html = (f"<b>ID de Escolha: {station_id_display} (Use o pop-up)</b><br>"
                          f"Nome: {station['name']}<br>"
                          f"ID Oficial: {station['id']}")
            
            folium.Marker(
                location=[station['latitude'], station['longitude']],
                popup=folium.Popup(popup_html, max_width=300),
                tooltip=f"Escolha #{station_id_display}"
            ).add_to(m)
        
        map_file = 'stations_map.html' # Salva na raiz do projeto
        m.save(map_file)
        return map_file

    def go_to_start_page(self):
        """Importa e navega para a página inicial."""
        from ..start_page import StartPage
        self.controller.show_frame(StartPage)
        
    def update_texts(self):
        """Atualiza os textos APENAS para esta tela."""
        i18n = self.controller.i18n
        self.controller.title(i18n.get('app_main_title'))
        self.back_button.config(text=i18n.get('back_btn'))
        self.page_title.config(text=i18n.get('download_data_title'))
        self.method_label.config(text=i18n.get('download_method_label'))
        self.city_label.config(text=i18n.get('city_label'))
        self.period_frame.config(text=i18n.get('time_period_label'))
        self.start_label.config(text=i18n.get('start_label'))
        self.end_label.config(text=i18n.get('end_label'))
        self.radius_label.config(text=i18n.get('radius_label'))
        self.download_btn.config(text=i18n.get('start_download_btn'))
        self.method_hint.text = i18n.get('download_method_hint')
        self.city_hint.text = i18n.get('city_download_entry_hint')
        self.start_hint.text = i18n.get('start_date_download_hint')
        self.end_hint.text = i18n.get('end_date_download_hint')
        self.radius_hint = i18n.get('radius_download_hint')


import tkinter as tk
from tkinter import ttk
from tkinter import Label, LabelFrame, Toplevel, StringVar, IntVar, Entry, Button, Checkbutton, CENTER
from tksheet import Sheet
import threading
from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from ..meta_learning.meta_learning import MetaLearning
from ..styles import colors

class TestsGenerator(Toplevel):
    # antiga MetaLearning
    def __init__(self, master=None):
        Toplevel.__init__(self, master=master)
        self.title('Meta Learning')
        self.geometry("800x800")

        self.configure(background=colors.fundo)

        Label(self, text='META-LEARNING', font='Arial 14 bold', fg='white', bg=colors.fundo).place(x=240, y=20)

        LabelFrame(self, text='CUSTOM TEST:', width=600, height=450, font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=20, y=60)

        Label(self, text='Base-Learning (Level 0):', font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=40, y=90)
        self.ml_lv0_p = StringVar()
        self.ml_lv0_p.set('Decision Trees')
        lista_ml0 =  ['None','Decision Trees', 'Bagged Trees', 'Neural network', 'Nearest Neighbors', 'Support Vector', 'Gaussian Process']
        ttk.Combobox(self, values=lista_ml0, textvariable=self.ml_lv0_p, width=30, font='Arial 11', justify=CENTER, state='readonly').place(x=40, y=120)

        Label(self, text='Triangulation (Level 0):', font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=40, y=190)
        self.ml_tr0_p = StringVar()
        self.ml_tr0_p.set('Arithmetic Average')
        lista_tr0 =  ['None', 'Arithmetic Average', 'Inverse Distance Weighted', 'Regional Weight', 'Optimized Normal Ratio']
        ttk.Combobox(self, values=lista_tr0, textvariable=self.ml_tr0_p, width=30, font='Arial 11', justify=CENTER, state='readonly').place(x=40, y=220)

        Label(self, text='Meta-Learning (Level 1):', font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=340, y=145)
        self.ml_lv1 = StringVar()
        self.ml_lv1.set('Decision Trees')
        lista_ml1 =  ['Decision Trees', 'Bagged Trees', 'Neural network', 'Nearest Neighbors', 'Support Vector', 'Gaussian Process']
        ttk.Combobox(self, values=lista_ml1, textvariable=self.ml_lv1, width=30, font='Arial 11', justify=CENTER, state='readonly').place(x=340, y=175)

        Label(self, text='Indicador climático:', font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=40, y=270)
        self.ind_meta_perso = StringVar()
        self.ind_meta_perso.set('Maximum temperature')
        lista_ind_meta_p = ["Precipitation", 'Maximum temperature', 'Minimum temperature']
        ttk.Combobox(self, values=lista_ind_meta_p, textvariable=self.ind_meta_perso, width=30, font='Arial 11', justify=CENTER, state='readonly').place(x=40, y=300)

        self.num_teste_mtp = IntVar()
        self.num_teste_mtp.set(1)
        Label(self, text="Number of tests (int):", font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=340, y=270)
        self.ent_num_teste = Entry(self, textvariable=self.num_teste_mtp, width=29, font='Arial 12', justify=CENTER).place(x=340, y=300)

        Label(self, text='Would you like to use any pre-configured ML model?', font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=40, y=340)
        self.pre_para_lv0 = IntVar()
        self.pre_para_lv1 = IntVar()
        Checkbutton(self, text='Level 0', variable=self.pre_para_lv0, bg=colors.fundo, font='Arial 12 bold', activebackground=colors.fundo).place(x=40, y=360)
        Checkbutton(self, text='Level 1', variable=self.pre_para_lv1, bg=colors.fundo, font='Arial 12 bold', activebackground=colors.fundo).place(x=180, y=360)
        
        Label(self, text='Would you like to use sliding window?', font='Arial 12 bold', fg='White', bg=colors.fundo).place(x=40, y=400)
        self.type_input = StringVar()
        self.type_input.set('Yes')
        lista_type_input = ['Yes', 'No']
        ttk.Combobox(self, values=lista_type_input, textvariable=self.type_input, width=30, font='Arial 11', justify=CENTER, state='readonly').place(x=40, y=430)
        
        self.button_ct = Button(self, text='Generate Preview', font='Arial 11 bold', bg=colors.fun_meta_le, fg='white', width=62, command=self.on_click_ct)
        self.button_ct.place(x=40, y=470)

        '''Teste global'''
        LabelFrame(self, text='GLOBAL TEST:', width=600, height=210, font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=20, y=520)
        
        Label(self, text='Wich MLs woud you like to use?', font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=40, y=550)

        self.pre_nn_comb = IntVar()
        self.pre_dt_comb = IntVar()
        self.pre_bt_comb = IntVar()
        self.pre_nneig_comb = IntVar()
        self.pre_sv_comb = IntVar()
        self.pre_gp_comb = IntVar()
        Checkbutton(self, text='NN', variable=self.pre_nn_comb, bg=colors.fundo, font='Arial 12 bold', activebackground=colors.fundo).place(x=40, y=580)
        Checkbutton(self, text='DT', variable=self.pre_dt_comb, bg=colors.fundo, font='Arial 12 bold', activebackground=colors.fundo).place(x=140, y=580)
        Checkbutton(self, text='NNeig.', variable=self.pre_nneig_comb, bg=colors.fundo, font='Arial 12 bold', activebackground=colors.fundo).place(x=240, y=580)
        Checkbutton(self, text='SV', variable=self.pre_sv_comb, bg=colors.fundo, font='Arial 12 bold', activebackground=colors.fundo).place(x=340, y=580)
        Checkbutton(self, text='GP', variable=self.pre_gp_comb, bg=colors.fundo, font='Arial 12 bold', activebackground=colors.fundo).place(x=440, y=580)
        Checkbutton(self, text='BT.', variable=self.pre_bt_comb, bg=colors.fundo, font='Arial 12 bold', activebackground=colors.fundo).place(x=540, y=580)

        Label(self, text='Climatic Indicator:', font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=40, y=620)
        self.ind_meta_comb = StringVar()
        self.ind_meta_comb.set('Maximum temperature')
        lista_ind_meta_p = ["Precipitation", 'Maximum temperature', 'Minimum temperature']
        ttk.Combobox(self, values=lista_ind_meta_p, textvariable=self.ind_meta_comb, width=30, font='Arial 11', justify=CENTER, state='readonly').place(x=40, y=650)

        self.num_teste_mtc = IntVar()
        self.num_teste_mtc.set(1)
        Label(self, text="Number of tests (int):", font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=340, y=620)
        self.ent_num_teste = Entry(self, textvariable=self.num_teste_mtc, width=29, font='Arial 12', justify=CENTER).place(x=340, y=650)
        
        self.button_gt = Button(self, text='Generate Preview GT', font='Arial 11 bold', bg=colors.fun_meta_le, fg='white', width=62, command=self.on_click_gt)
        self.button_gt.place(x=40, y=690)

        #Aviso
        LabelFrame(self, text='Warning!:', width=600, height=120, font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=20, y=740)
        Label(self, text='Depending on the combination selected in the "Custom Test" or in the "GLOBAL ', font='Arial 12 bold', fg='#FF8C00', bg=colors.fundo).place(x=40, y=770)
        Label(self, text='TEST", it may take a few minutes due to processing', font='Arial 12 bold', fg='#FF8C00', bg=colors.fundo).place(x=40, y=790)
        Label(self, text='Feel free to use your computer for other tasks in meantime.', font='Arial 12 bold', fg='#FF8C00', bg=colors.fundo).place(x=40, y=820)

    def on_click_gt(self):
        self.button_gt.config(command=None)  # Remove o comando temporariamente
        self.button_gt.config(fg='white')    # Força a cor branca
        self.loading = True
        self.loading_step = 0
        self.animate_loading_gt()
        threading.Thread(target=self.run_process_gt).start()

    def animate_loading_gt(self):
        if self.loading:
            dots = '.' * (self.loading_step % 4)
            self.button_gt.config(text=f"Loading{dots}")
            self.loading_step += 1
            self.after(500, self.animate_loading_gt)

    def run_process_gt(self):
        self.generate_global_test()
        self.after(0, self.reset_button_gt)

    def reset_button_gt(self):
        self.loading = False
        self.button_gt.config(text="Generate Preview GT", command=self.on_click_gt)

    def on_click_ct(self):
        self.button_ct.config(command=None)  # Remove o comando temporariamente
        self.button_ct.config(fg='white')    # Força a cor branca
        self.loading = True
        self.loading_step = 0
        self.animate_loading_ct()
        threading.Thread(target=self.run_process_ct).start()

    def animate_loading_ct(self):
        if self.loading:
            dots = '.' * (self.loading_step % 4)
            self.button_ct.config(text=f"Loading{dots}")
            self.loading_step += 1
            self.after(500, self.animate_loading_ct)

    def run_process_ct(self):
        self.generate_custom_test()
        self.after(0, self.reset_button_ct)

    def reset_button_ct(self):
        self.loading = False
        self.button_ct.config(text="Generate Preview", command=self.on_click_ct)

    def generate_custom_test(self):
        """
        Antiga gerar_teste_perso
        """
        # Retrieve user inputs
        base_model = self.ml_lv0_p.get()
        triangulation = self.ml_tr0_p.get()
        meta_model = self.ml_lv1.get()
        indicator = self.ind_meta_perso.get()
        num_tests = int(self.num_teste_mtp.get())
        pre_level_0 = int(self.pre_para_lv0.get())
        pre_level_1 = int(self.pre_para_lv1.get())
        input_window = self.type_input.get()

        # Determine focus based on indicator type
        if indicator == "Precipitation":
            focus = 1
        elif indicator == 'Maximum temperature':
            focus = 2
        elif indicator == 'Minimum temperature':
            focus = 3

        # Perform meta-learning
        meta = MetaLearning()
        meta_ea, meta_er, meta_percent_error, meta_r2, x_meta, y_meta, y_target, \
        base_ea, base_er, base_percent, base_r2, tria_ea, tria_er = \
            meta.customized_meta_learning(focus, base_model, triangulation, meta_model, 0, 0, num_tests, input_window)

        # Create result preview frame
        LabelFrame(self, text='RESULTS PREVIEW:', width=1250, height=950,
                font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=640, y=60)

        # Display error metrics
        errors = [
            ("ABSOLUTE ERROR:", base_ea, tria_ea, meta_ea),
            ("RELATIVE ERROR:", base_er, tria_er, meta_er),
            ("ERROR (%):", base_percent, tria_ea * 100, meta_percent_error),
            ("R2:", base_r2, None, meta_r2)
        ]

        y_positions = [90, 130, 170, 210]
        for idx, (label, val1, val2, val3) in enumerate(errors):
            text = f"{label:<20}  Machine Learning: {round(val1, 4)}"
            if val2 is not None:
                text += f"   ||   Triangulation: {round(val2, 4)}"
            if val3 is not None:
                text += f"   ||   Meta Learning: {round(val3, 4)}"
            Label(self, text=text, font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=660, y=y_positions[idx])

        # Create figure for plotting results
        figure = Figure(figsize=(12.3, 7.5), dpi=100)

        # Absolute Error
        ax1 = figure.add_subplot(2, 2, 1)
        ax1.bar(["ML", "Triang", "Meta"], [base_ea, tria_ea, meta_ea])
        ax1.set_ylabel("Absolute Error")

        # Relative Error
        ax2 = figure.add_subplot(2, 2, 2)
        ax2.bar(["ML", "Triang", "Meta"], [base_er, tria_er, meta_er])
        ax2.set_ylabel("Relative Error")

        # Percentage Error
        ax3 = figure.add_subplot(2, 2, 3)
        ax3.bar(["ML", "Triang", "Meta"], [base_percent, tria_ea * 100, meta_percent_error])
        ax3.set_ylabel("Error (%)")

        # R² Score
        ax4 = figure.add_subplot(2, 2, 4)
        ax4.bar(["ML", "Meta"], [base_r2, meta_r2])
        ax4.set_ylabel("R²")

        # Embed the figure in the Tkinter window
        canvas = FigureCanvasTkAgg(figure, master=self)
        canvas.draw()
        canvas.get_tk_widget().pack()
        canvas.get_tk_widget().place(x=650, y=250)

        # Add navigation toolbar
        toolbar = NavigationToolbar2Tk(canvas, self)
        toolbar.place(x=1150, y=10)
        toolbar.update()

    def generate_global_test(self):
        target_variable = self.ind_meta_comb.get()
        num_tests = int(self.num_teste_mtc.get())
        window_type = 'Yes'

        meta_learner = MetaLearning()
        all_models, model_ranking = meta_learner.combine_meta_learning(target_variable, 0, 0, num_tests, window_type)

        LabelFrame(
            self, text='RESULTS:', width=1260, height=950,
            font='Arial 12 bold', fg='white', bg=colors.fundo
        ).place(x=640, y=60)

        Label(
            self, text='Generated Models:', font='Arial 12 bold',
            fg='white', bg=colors.fundo
        ).place(x=660, y=90)

        all_data = []
        for model in all_models:
            row = [
                model[0],  # Model name
                model[1],  # Base learning
                model[2],  # Triangulation
                model[3],  # Meta learning
                model[5],  # Absolute Error
                model[6],  # Relative Error
                model[7],  # Percentage Error
            ]
            all_data.append(row)

        self.all_models_table = Sheet(
            self, data=all_data,
            headers=[
                'Model', 'Base Learning', 'Triangulation',
                'Meta Learning', 'Absolute Error',
                'Relative Error', 'Error (%)'
            ],
            width=890, height=500
        )
        self.all_models_table.enable_bindings()
        self.all_models_table.place(x=660, y=120)

        Label(
            self, text='Model Ranking:', font='Arial 12 bold',
            fg='white', bg=colors.fundo
        ).place(x=1580, y=90)

        ranking_data = []
        x_labels = []
        y_values = []
        for i, model in enumerate(model_ranking):
            model_name = model[0]
            error_value = round(float(model[1].replace(',', '.')), 4)
            ranking_data.append([model_name, error_value])

            if i <= 15:
                x_labels.append(str(model_name))
                y_values.append(error_value)

        self.ranking_table = Sheet(
            self, data=ranking_data,
            headers=['Model', 'Error'],
            width=270, height=500, column_width=115
        )
        self.ranking_table.enable_bindings()
        self.ranking_table.place(x=1580, y=120)

        figure = Figure(figsize=(12, 3.3), dpi=100)
        plot = figure.add_subplot(1, 1, 1)

        plot.bar(x_labels, y_values)
        plot.set_ylabel('Error (%)')
        plot.set_xlabel('Models')
        plot.grid(True)

        canvas = FigureCanvasTkAgg(figure, master=self)
        canvas.draw()
        canvas.get_tk_widget().pack()
        canvas.get_tk_widget().place(x=660, y=650)

        toolbar = NavigationToolbar2Tk(canvas, self)
        toolbar.place(x=1150, y=10)
        toolbar.update()

     



from .utils import *
from .tooltip import CreateToolTip

if typing.TYPE_CHECKING:
    from .data_imputation.data_imputation_page import DataImputationPage
    from .download.download_page import DownloadDataPage
    from .tutorial.tutorial_page import TutorialPage
    from .trends.trends_page import ClimateTrendsPage

class StartPage(ttk.Frame):
    def __init__(self, parent, controller):
        ttk.Frame.__init__(self, parent)
        self.controller = controller

        # --- Widgets ---
        self.welcome_label = ttk.Label(self, text="", font=("Verdana", 20, "bold"))
        self.welcome_label.pack(pady=40, padx=10)

        button_frame = ttk.Frame(self)
        button_frame.pack()

        i18n = self.controller.i18n

        self.tutorial_btn = ttk.Button(button_frame, text="", command=self.go_to_tutorial)
        self.tutorial_btn.pack(pady=10, fill=tk.X)
        self.tutorial_tooltip = CreateToolTip(self.tutorial_btn, text=i18n.get('tutorial_home_hint'))

        self.download_btn = ttk.Button(button_frame, text="", command=self.go_to_download)
        self.download_btn.pack(pady=10, fill=tk.X)
        self.download_tooltip = CreateToolTip(self.download_btn, text=i18n.get('download_home_hint'))

        self.imputation_btn = ttk.Button(button_frame, text="", command=self.go_to_data_imputation)
        self.imputation_btn.pack(pady=10, fill=tk.X)
        self.imputation_tooltip = CreateToolTip(self.imputation_btn, text=i18n.get('imputation_home_hint'))

        self.trends_btn = ttk.Button(button_frame, text="", command=self.go_to_trends)
        self.trends_btn.pack(pady=10, fill=tk.X)
        self.trends_tooltip = CreateToolTip(self.trends_btn, text=i18n.get('trends_home_hint'))

        # --- Seletor de Idioma ---
        lang_frame = ttk.Frame(self)
        lang_frame.pack(pady=50)
        self.lang_label = ttk.Label(lang_frame, text="")
        self.lang_label.pack(side=tk.LEFT, padx=5)

        self.lang_combobox = ttk.Combobox(lang_frame, values=list(controller.i18n.available_languages.keys()), state="readonly")
        self.lang_combobox.pack(side=tk.LEFT)
        self.lang_combobox.set("Português (BR)")
        self.lang_combobox.bind("<<ComboboxSelected>>", self.change_language)
        
        # Inicializa os textos
        self.update_texts()

    def go_to_data_imputation(self):
        """Importa e navega para a página de imputação de dados."""
        from .data_imputation.data_imputation_page import DataImputationPage
        self.controller.show_frame(DataImputationPage)

    def go_to_tutorial(self):
        """Importa e navega para a página de tutorial."""
        from .tutorial.tutorial_page import TutorialPage
        self.controller.show_frame(TutorialPage)

    def go_to_download(self):
        """Importa e navega para a página de download de dados."""
        from .download.download_page import DownloadDataPage
        self.controller.show_frame(DownloadDataPage)

    def go_to_trends(self):
        """Importa e navega para a página de tendencias climaticas."""
        from .trends.trends_page import ClimateTrendsPage
        self.controller.show_frame(ClimateTrendsPage)

    def update_texts(self):
        """Atualiza os textos APENAS para esta tela."""
        i18n = self.controller.i18n
        self.welcome_label.config(text=i18n.get('welcome_text'))
        self.tutorial_btn.config(text=i18n.get('tutorial_btn'))
        self.download_btn.config(text=i18n.get('download_data_btn'))
        self.imputation_btn.config(text=i18n.get('data_imputation_btn'))
        self.trends_btn.config(text=i18n.get('climatic_trends_btn'))
        self.lang_label.config(text=i18n.get('language_label'))
        self.tutorial_tooltip.text = i18n.get('tutorial_home_hint')
        self.download_tooltip.text = i18n.get('download_home_hint')
        self.imputation_tooltip.text = i18n.get('imputation_home_hint')
        self.trends_tooltip.text = i18n.get('trends_home_hint')

    def change_language(self, event=None):
        selected_display_name = self.lang_combobox.get()
        lang_code = self.controller.i18n.available_languages[selected_display_name]
        
        self.controller.i18n.set_language(lang_code)
        self.controller.update_all_frames_text()




import numpy as np

def calculate_errors(all_exact, all_predicted, total_absolute_error, total_relative_error, n_tests):
    score = round((((total_relative_error / n_tests) * 100) - 100) * -1, 2)

    if all_exact:
        errors = np.abs(np.array(all_exact) - np.array(all_predicted))
        max_idx = np.argmax(errors)
        min_idx = np.argmin(errors[errors > 0]) if np.any(errors > 0) else 0

        max_error = errors[max_idx]
        min_error = errors[min_idx] if np.any(errors > 0) else 0

        exact_max = all_exact[max_idx]
        predicted_max = all_predicted[max_idx]

        exact_min = all_exact[min_idx]
        predicted_min = all_predicted[min_idx]
    else:
        max_error = min_error = exact_max = predicted_max = exact_min = predicted_min = 0

    avg_absolute_error = total_absolute_error / n_tests
    avg_relative_error = total_relative_error / n_tests

    return {
        "score": score,
        "max_error": max_error,
        "min_error": min_error,
        "exact_max": exact_max,
        "predicted_max": predicted_max,
        "exact_min": exact_min,
        "predicted_min": predicted_min,
        "avg_absolute_error": avg_absolute_error,
        "avg_relative_error": avg_relative_error
    }



from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from ..meta_learning.meta_learning import MetaLearning

class TestsGenerator:
    def __init__(self):
        self.ml_lv0_p = 'Decision Trees'
        lista_ml0 =  ['None','Decision Trees', 'Bagged Trees', 'Neural network', 'Nearest Neighbors', 'Support Vector', 'Gaussian Process']
        self.ml_tr0_p = 'Arithmetic Average'
        lista_tr0 =  ['None', 'Arithmetic Average', 'Inverse Distance Weighted', 'Regional Weight', 'Optimized Normal Ratio']
        self.ml_lv1 = 'Decision Trees'
        lista_ml1 =  ['Decision Trees', 'Neural network', 'Bagged Trees', 'Nearest Neighbors', 'Support Vector', 'Gaussian Process']
        self.ind_meta_perso = 'Maximum temperature'
        lista_ind_meta_p = ["Precipitation", 'Maximum temperature', 'Minimum temperature']
        self.num_teste_mtp = 1
        self.pre_para_lv0 = False
        self.pre_para_lv1 = False
        # sliding window
        self.type_input = 'Yes'

        self.pre_nn_comb = False
        self.pre_dt_comb = False
        self.pre_nneig_comb = False
        self.pre_sv_comb = False
        self.pre_gp_comb = False
        self.pre_bt_comb = False

        self.ind_meta_comb = 'Maximum temperature'
        self.num_teste_mtc = 1
        

    def generate_custom_test(self, base_model='Decision Trees', triangulation='Arithmetic Average',
                             meta_model='Decision Trees', indicator='Maximum temperature',
                             num_tests=1, input_window='Yes'):

        if indicator == "Precipitation":
            focus = 1
        elif indicator == 'Maximum temperature':
            focus = 2
        elif indicator == 'Minimum temperature':
            focus = 3

        # Perform meta-learning
        meta = MetaLearning()
        meta_ea, meta_er, meta_percent_error, meta_r2, x_meta, y_meta, y_target, \
        base_ea, base_er, base_percent, base_r2, tria_ea, tria_er = \
            meta.customized_meta_learning(focus, base_model, triangulation, meta_model, 0, 0, num_tests, input_window)
        
        # Mostra resultados no terminal
        print("\n===== RESULTS PREVIEW =====")
        print(f"ABSOLUTE ERROR:  ML: {round(base_ea, 4)}  ||  Triangulation: {round(tria_ea, 4)}  ||  Meta: {round(meta_ea, 4)}")
        print(f"RELATIVE ERROR:  ML: {round(base_er, 4)}  ||  Triangulation: {round(tria_er, 4)}  ||  Meta: {round(meta_er, 4)}")
        print(f"ERROR (%):       ML: {round(base_percent, 4)}  ||  Triangulation: {round(tria_ea * 100, 4)}  ||  Meta: {round(meta_percent_error, 4)}")
        print(f"R²:              ML: {round(base_r2, 4)}  ||  Meta: {round(meta_r2, 4)}")

        # Plota os gráficos com matplotlib
        fig, axs = plt.subplots(2, 2, figsize=(12, 8))

        axs[0, 0].bar(["ML", "Triang", "Meta"], [base_ea, tria_ea, meta_ea])
        axs[0, 0].set_title("Absolute Error")
        axs[0, 0].set_ylabel("Error")

        axs[0, 1].bar(["ML", "Triang", "Meta"], [base_er, tria_er, meta_er])
        axs[0, 1].set_title("Relative Error")
        axs[0, 1].set_ylabel("Error")

        axs[1, 0].bar(["ML", "Triang", "Meta"], [base_percent, tria_ea * 100, meta_percent_error])
        axs[1, 0].set_title("Percentage Error")
        axs[1, 0].set_ylabel("Error (%)")

        axs[1, 1].bar(["ML", "Meta"], [base_r2, meta_r2])
        axs[1, 1].set_title("R² Score")
        axs[1, 1].set_ylabel("Score")

        plt.tight_layout()
        plt.show()

    def generate_global_test(self, indicator='Maximum temperature', 
                             num_tests=1, window_type='Yes', print_results=True
    ):
        """
        Versão sem Tkinter da função generate_global_test
        """
        meta_learner = MetaLearning()
        # Executa o algoritmo de meta-aprendizado
        all_models, model_ranking = meta_learner.combine_meta_learning(
            indicator, 0, 0, num_tests, window_type
        )

        if print_results:
            print("=== GENERATED MODELS ===")
            headers = [
                'Model', 'Base Learning', 'Triangulation',
                'Meta Learning', 'Absolute Error',
                'Relative Error', 'Error (%)'
            ]
            print("\t".join(headers))
            for model in all_models:
                row = [
                    f"{model[0]}",  # Model name
                    model[1],  # Base learning
                    model[2],  # Triangulation
                    model[3],  # Meta learning
                    f"{model[5]:.4f}",  # Absolute Error
                    f"{model[6]:.4f}",  # Relative Error
                    f"{model[7]:.4f}"   # Percentage Error
                ]
                print("\t".join(row))

            print("\n=== MODEL RANKING ===")
            print("Model\tError (%)")
            for model in model_ranking:
                print(f"{model[0]}\t{model[1]}")

        # Preparar dados para gráfico
        x_labels = []
        y_values = []

        for i, model in enumerate(model_ranking[:15]):
            model_name = model[0]
            error_value = float(str(model[1]).replace(',', '.'))
            x_labels.append(model_name)
            y_values.append(error_value)

        # Criar gráfico com matplotlib puro
        plt.figure(figsize=(12, 3.3))
        plt.bar(x_labels, y_values)
        plt.ylabel("Error (%)")
        plt.xlabel("Models")
        plt.title("Model Ranking - Top 15")
        plt.grid(True)
        plt.xticks(rotation=45)
        plt.tight_layout()
        plt.show()

        return all_models, model_ranking




from tkinter import Canvas, Label, LabelFrame, StringVar, IntVar, BooleanVar, Entry, Scale, Checkbutton, Button, HORIZONTAL, CENTER
from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from ..training.training import Training
from ..styles import colors
from ..utils.indicator import get_indicator_code
from ..utils.ml_param import get_parameters_ml
from ..utils.insert_canvas import insert_canvas_toolbar

class View:
    def data_preview(self, pts, media_ea, media_er, maior_ea, exat_maior, pre_maior, menor_ea, exat_menor, pre_menor, eixo_y_exato, eixo_y_predict, eixo_x):
        self.laf_res = LabelFrame(self, text='Preview dos resultados', width=1250, height=950, font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=650, y=50)
        Label(self, text='Pontuação (0-100): '+ str(pts) +'pts', font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=680, y=70)
        media_ea = round(media_ea, 4)
        Label(self, text='Média Erro absoluto: '+ str(media_ea), font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=680, y=100)
        media_er = round(media_er, 4)
        Label(self, text='Média Erro relativo: '+ str(media_er), font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=680, y=130)

        Label(self, text='Maior erro absoluto: ' + str(round(maior_ea,4)), font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=680, y=160)
        Label(self, text="Valor exato do maior EA: " + str(round(exat_maior,4)),font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=940, y=160)
        Label(self, text="Predict do maior EA: " + str(round(pre_maior, 4)), font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=1200, y=160)

        Label(self, text='Menor erro absoluto: ' + str(round(menor_ea,4)), font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=680, y=190)
        Label(self, text="Valor exato do menor EA: " + str(round(exat_menor,4)),font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=940, y=190)
        Label(self, text="Predict do menor EA: " + str(round(pre_menor, 4)), font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=1200, y=190)

        figura = Figure(figsize=(12,7.3), dpi=100)
        plot_r = figura.add_subplot(111)
        plot_r.plot(eixo_x, eixo_y_exato,label='Exato', color='green')
        plot_r.plot(eixo_x, eixo_y_predict, label='Predict', color='red')
        plot_r.legend()
        plot_r.grid(True)
        plot_r.set_ylabel("Temperature(°C)")
        plot_r.set_xlabel("Comparações")

        canvas = FigureCanvasTkAgg(figura, master=self)
        insert_canvas_toolbar(canvas, self)

    def generate_preview_dt(self):
        prev = Training()
        salvar_m = self.save_model.get()
        
        #cidade = self.get_end(self.data_s.get())
        cidade = self.data_s.get()
        indicator = self.ind_s.get()
        divisao = int(self.por_trei.get())
        criterio = self.criterion_v.get()
        splitter = self.splitter_v.get()
        maxd = int(self.maxd_v.get())             #* Max_depth
        minsams = self.int_float(self.minsam_s_v.get())    #* Min_samples_split
        minsaml = self.int_float(self.minsam_l_v.get())    #* Min_samples_leaf
        minwei = float(self.minweifra_l_v.get())
        maxfe = self.valid_maxf(self.maxfeat_v.get())
        maxleaf = int(self.maxleaf_n.get())
        
        minim = float(self.minimp_dec.get())
        ccp = float(self.ccp_alp_v.get())
        n_tes = int(self.num_teste.get())

        indicator = get_indicator_code(indicator)

        pts, media_ea, media_er, maior_ea, exat_maior, pre_maior, menor_ea, exat_menor, pre_menor, eixo_y_exato, eixo_y_predict, eixo_x = prev.decision_tree(cidade, indicator, divisao, criterio, splitter, maxd, minsaml, maxfe, maxleaf, n_tes, minsams, minwei, minim, ccp, salvar_m)
        self.data_prev(pts, media_ea, media_er, maior_ea, exat_maior, pre_maior, menor_ea, exat_menor, pre_menor, eixo_y_exato, eixo_y_predict, eixo_x)

    def generate_preview_bt(self):
        prev = Training()
        salvar_m = self.save_model.get()
        
        #cidade = self.get_end(self.data_s.get())
        cidade = self.data_s.get()
        indicator = self.ind_s.get()
        divisao = int(self.por_trei.get())
        criterio = self.criterion_v.get()
        splitter = self.splitter_v.get()
        maxd = int(self.maxd_v.get())             #* Max_depth
        minsams = self.int_float(self.minsam_s_v.get())    #* Min_samples_split
        minsaml = self.int_float(self.minsam_l_v.get())    #* Min_samples_leaf
        minwei = float(self.minweifra_l_v.get())
        maxfe = self.valid_maxf(self.maxfeat_v.get())
        maxleaf = int(self.maxleaf_n.get())
        
        minim = float(self.minimp_dec.get())
        ccp = float(self.ccp_alp_v.get())
        n_tes = int(self.num_teste.get())
        n_estimators = int(self.n_estimators.get())

        indicator = get_indicator_code(indicator)

        pts, media_ea, media_er, maior_ea, exat_maior, pre_maior, menor_ea, exat_menor, pre_menor, eixo_y_exato, eixo_y_predict, eixo_x = prev.bagging_trees(cidade, indicator, divisao, criterio, splitter, maxd, minsaml, maxfe, maxleaf, n_tes, minsams, minwei, minim, ccp, salvar_m, n_estimators)
        self.data_prev(pts, media_ea, media_er, maior_ea, exat_maior, pre_maior, menor_ea, exat_menor, pre_menor, eixo_y_exato, eixo_y_predict, eixo_x)


    def generate_preview_nn(self):
        prev = Training()
        salvar_m = self.save_model.get()
        
        cidade = self.get_end(self.data_s.get())
        
        indicator = self.ind_s.get()
        indicator = get_indicator_code(indicator)
    
        divisao = int(self.por_trei.get())

        activ = self.activation_v.get()
        solv = self.solver_v.get()
        alph = float(self.alpha_v.get())
        batc = self.batch_size_v.get()
        learn_r = self.learning_rate_v.get()
        learn_r_ini = float(self.learning_rate_init_v.get())
        powt = float(self.power_t_v.get())
        maxit = int(self.max_iter_v.get())
        shuf = self.shuffle_v.get()
        tol = float(self.tol_v.get())
        verb = self.verbose_v.get()
        warms = self.warm_start_v.get()
        moment = float(self.momentum_v.get())
        neste = self.nesterovs_momentum_v.get()
        earlyst = self.early_stopping_v.get()
        valid = float(self.validation_fraction_v.get())
        b1 = float(self.beta_1_v.get())
        b2 = float(self.beta_2_v.get())
        niter = int(self.n_iter_no_change_v.get())
        maxfun = int(self.max_fun_v.get())
        n_teste = int(self.num_teste.get())
        pts, media_ea, media_er, maior_ea, exat_maior, pre_maior, menor_ea, exat_menor, pre_menor, eixo_y_exato, eixo_y_predict, eixo_x = prev.RedeNeural(cidade, indicator, divisao, n_teste, activ, solv, alph, batc, learn_r, learn_r_ini, powt, maxit, shuf, tol, verb, warms, moment, neste, earlyst, valid, b1, b2, niter, maxfun, salvar_m)

        self.data_prev(pts, media_ea, media_er, maior_ea, exat_maior, pre_maior, menor_ea, exat_menor, pre_menor, eixo_y_exato, eixo_y_predict, eixo_x)
  
    def generate_preview_svm(self):
        prev = Training()
        salvar_m = self.save_model.get()
        
        cidade = self.get_end(self.data_s.get())
        
        indicator = self.ind_s.get()
        indicator = get_indicator_code(indicator)
    
        divisao = int(self.por_trei.get())
        n_teste = int(self.num_teste.get())
        kern = self.kernel_v.get()
        degre = self.degree_v.get()
        gam = self.gamma_v.get()
        coef = float(self.coef0_v.get())
        t = float(self.tol_v.get())
        c = float(self.c_v.get())
        eps = float(self.epsilon_v.get())
        shr = self.shrinking_v.get()
        cach = float(self.cache_size_v.get())
        verb = self.verbose_v.get()
        maxi = int(self.maxiter_v.get())


        pts, media_ea, media_er, maior_ea, exat_maior, pre_maior, menor_ea, exat_menor, pre_menor, eixo_y_exato, eixo_y_predict, eixo_x = prev.SVR(cidade, indicator, divisao, n_teste, kern, degre, gam, coef, t, c, eps, shr, cach, verb, maxi, salvar_m)

        self.data_prev(pts, media_ea, media_er, maior_ea, exat_maior, pre_maior, menor_ea, exat_menor, pre_menor, eixo_y_exato, eixo_y_predict, eixo_x)
      
    def generate_preview_Kn(self):
        prev = Training()
        parameters = get_parameters_ml(self)

        cidade = parameters["cidade"]
        salvar_m = parameters["save_model"]
        n_tes = parameters["num_testes"]
        divisao = parameters["porcentagem_treinamento"]
        n_neig = parameters["n_neighbors"]
        algor = parameters["algorithm"]
        leaf_s = parameters["leaf_size"]
        pv = parameters["p_v"]
        n_job = parameters["n_jobs"]
            
        indicator = self.ind_s.get()
        indicator = get_indicator_code(indicator)

        pts, media_ea, media_er, maior_ea, exat_maior, pre_maior, menor_ea, exat_menor, pre_menor, eixo_y_exato, eixo_y_predict, eixo_x = \
            prev.KNeighbors(cidade, indicator, divisao, n_tes, n_neig, algor, leaf_s, pv, n_job, salvar_m)
        self.data_prev(pts, media_ea, media_er, maior_ea, exat_maior, pre_maior, menor_ea, exat_menor, pre_menor, eixo_y_exato, eixo_y_predict, eixo_x)



"""Module for the machine learning without interface.

This module creates a window with options for selecting different 
Machine Learning algorithms and integrates functionalities such as
parameter selection, data preview, and result visualization.
"""

from .ml_view import View
from .utils import *
from .ml_helpers import *
from ..data_processing.data_processing import DataProcessing
import optuna
import threading
import json
import functools
import os
import sys
from ..utils.indicator import get_indicator_code 
from ..training.training import Training

class Ml:
    def __init__(self):
        self.list_ml = ['Decision Trees',
                        'Bagged Trees',
                        'Neural network',
                        'Nearest Neighbors',
                        'Support Vector',
                        'Gaussian Process']
        
        self.list_ind = ["Precipitation", 'Maximum temperature', 'Minimum temperature']
        self.list_dt = ['Target city', 'Neighbor A', 'Neighbor B', 'Neighbor C']
        
        self.data_s = "Target_city"
        self.ind_s = "Maximum temperature"
        self.last_best_params = {}
        

    def int_float(self, value):
        """
        Try to return a int value, if it can't, returns a float value
        """
        try:
            return int(value)
        except:
            return float(value)

    def valid_maxf(self, value):
        """
        Returnes the correct type of a received value
        """
        if value.isdigit() == True:
            value = int(value)
        elif value.isalnum() == True and value.isdigit() == False:
            value = str(value)
        elif value.isalnum() == False and value.isdigit() == False and value.isalpha() == False:
            value = float(value)

        return value

    def save_parameter(self):
        v = View
        return v.save_parameter()

    def data_preview(self, score, mean_abs_error, mean_rel_error, max_abs_error, exact_max,
                     pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis):
        v = View()
        v.data_preview(score, mean_abs_error, mean_rel_error, max_abs_error, exact_max,
                       pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis)

    def get_end(self, city):
        treatment = DataProcessing()
        return treatment.get_file_path(city)

    def generate_preview_dt(self):
        v = View()
        v.generate_preview_dt(self)
    
    def generate_preview_bt(self):
        v = View()
        v.generate_preview_bt(self)

    def generate_preview_nn(self):
        v = View()
        v.generate_preview_nn(self)

    def generate_preview_svm(self):
        v = View()
        v.generate_preview_svm(self)

    def generate_preview_kn(self):
        v = View()
        v.generate_preview_kn(self)

    def _load_json_params(self, load_params):
        try:
            with open(load_params, "r", encoding="utf-8") as file:
                data = json.load(file)
                return data.get("params", {})
        except Exception as e:
            print("Erro ao carregar JSON:", e)
            return None

    def decision_tree(self, criterion_v="squared_error", splitter="best", maxd_v='10',
                      minsam_s_v=2, minsam_l_v=50, minweifra_l_v='0.0', maxfeat_v="sqrt",
                      maxleaf_n='10', minimp_dec='0.0', ccp_alp_v='0.0', 
                      por_trei=70, num_teste=5,save_model=False, load_params='None'):
        
        if load_params is not 'None':
            try:
                with open(load_params, "r", encoding="utf-8") as file:
                    data = json.load(file)
                    criterion = data["criterion"]
                    splitter = data["splitter"]
                    max_depth = data["max_depth"]
                    min_samples_s = data["min_samples_split"]
                    min_samples_l = data["min_samples_leaf"]
                    max_features = data["max_features"]
                    min_impurity_d = data["min_impurity_decrease"]
                    ccp_alpha = data["ccp_alpha"]

                    self.param_frame = DTParameter(criterion, splitter, max_depth, min_samples_s,
                                                   min_samples_l, max_features, maxleaf_n, min_impurity_d, ccp_alpha)
                    
                    self.por_trei = por_trei
                    self.num_teste = num_teste
                    self.save_model = save_model
                    self.generate_preview_dt()
            except Exception as e:
                print("Ocorreu um erro:", e)
        
        list_cri = ["squared_error", "friedman_mse", "absolute_error", "poisson"]
        if criterion_v not in list_cri:
            raise ValueError(f"Selected criterion_v is not compatible. "
            f"Please try one of the following options: 'squared_error', 'friedman_mse', 'absolute_error', 'poisson'.")
        
        list_spl = ['best', 'random']
        if splitter not in list_spl:
            raise ValueError(f"Selected splitter is not compatible. "
                             f"Please try one of the following options: 'best', 'random'")
        
        list_maxfeat_v = ['int', 'float', 'sqrt', 'log2']
        if maxfeat_v not in list_maxfeat_v:
            raise ValueError(f"Selected maxfeat_v is not compatible. "
                             f"Please try one of the following options: 'int', 'float', 'sqrt', 'log2'")
        
        self.param_frame = DTParameter(criterion_v, splitter, maxd_v, minsam_s_v, minsam_l_v,
                                       minweifra_l_v, maxfeat_v, maxleaf_n, minimp_dec, ccp_alp_v)
        self.por_trei = por_trei
        self.num_teste = num_teste
        self.save_model = save_model
        self.generate_preview_dt()

    def bagged_trees(self, criterion_v="squared_error", splitter="best", maxd_v='10',
                      minsam_s_v=2, minsam_l_v=50, minweifra_l_v='0.0', maxfeat_v="sqrt",
                      maxleaf_n='10', minimp_dec='0.0', ccp_alp_v='0.0', 
                      por_trei=70, num_teste=5,save_model=False, n_estimators=10, load_params='None'):
        if load_params != 'None':
            params = self._load_json_params(load_params)
            if params:
                self.param_frame = BTParameter(
                    params["criterion"],
                    "best",
                    params["max_depth"],
                    params["min_samples_split"],
                    params["min_samples_leaf"],
                    '0.0',
                    params["max_features"],
                    maxleaf_n,
                    params["min_impurity_decrease"],
                    params["ccp_alpha"],
                    params["n_estimators_bag"]
                )
                
                self.por_trei = por_trei
                self.num_teste = num_teste
                self.save_model = save_model
                self.n_estimators = params["n_estimators_bag"]
                self.generate_preview_bt()
                return
        
        list_cri = ["squared_error", "friedman_mse", "absolute_error", "poisson"]
        if criterion_v not in list_cri:
            raise ValueError(f"Selected criterion_v is not compatible. "
            f"Please try one of the following options: 'squared_error', 'friedman_mse', 'absolute_error', 'poisson'.")
        
        list_spl = ['best', 'random']
        if splitter not in list_spl:
            raise ValueError(f"Selected splitter is not compatible. "
                             f"Please try one of the following options: 'best', 'random'")
        
        list_maxfeat_v = ['int', 'float', 'sqrt', 'log2']
        if maxfeat_v not in list_maxfeat_v:
            raise ValueError(f"Selected maxfeat_v is not compatible. "
                             f"Please try one of the following options: 'int', 'float', 'sqrt', 'log2'")
        
        self.param_frame = BTParameter(criterion_v, splitter, maxd_v, minsam_s_v, minsam_l_v,
                                       minweifra_l_v, maxfeat_v, maxleaf_n, minimp_dec, ccp_alp_v, n_estimators)
        self.por_trei = por_trei
        self.num_teste = num_teste
        self.save_model = save_model
        self.n_estimators = n_estimators
        self.generate_preview_bt()

    def neural_network(self, activation_v='relu', solver_v='adam', alpha_v='0.0001',
                       batch_size_v='auto', learning_rate_v='constant', learning_rate_init_v='0.001',
                       power_t_v='0.5', max_iter_v='200', shuffle_v=True, tol_v='0.0001',
                       verbose_v=False, warm_start_v=False, momentum_v='0.9', nesterovs_momentum_v=True,
                       early_stopping_v=False, validation_fraction_v='0.1', beta_1_v='0.9',
                       beta_2_v='0.999', n_iter_no_change_v='10', max_fun_v='15000', por_trei=70,
                       num_teste=5, save_model=False, load_params='None'):
        
        if load_params != 'None':
            params = self._load_json_params(load_params)
            if params:
                self.param_frame = NNParameter(
                    params["activation"],
                    params["solver"],
                    params["alpha"],
                    params["batch_size"],
                    params["learning_rate"],
                    params["learning_rate_init"],
                    '0.5',
                    params["max_iter"],
                    True, '0.0001', False, False,
                    params.get("momentum", 0.9),
                    True, False, '0.1',
                    0.9, 0.999, 10, 15000
                )
                self.save_model = save_model
                self.num_teste = num_teste
                self.por_trei = por_trei
                self.generate_preview_nn()
                return
        list_activation_v = ['identity', 'logistic', 'tanh', 'relu']
        if activation_v not in list_activation_v:
            raise ValueError(f"Selected activation_v is not compatible. "
                             f"Please try one of the following options: 'identity', 'logistic', 'tanh', 'relu'")
        
        list_solver = ['lbfgs', 'sgd', 'adam']
        if solver_v not in list_solver:
            raise ValueError(f"Selected solver_v is not compatible. "
                             f"Please try one of the following options: 'lbfgs', 'sgd', 'adam'")
        
        list_batch_size = ['int', 'auto']
        if batch_size_v not in list_batch_size:
            raise ValueError(f"Selected batch_size_v is not compatible. "
                             f"Please try one of the following options: 'int', 'auto'")
        
        list_learn = ['constant', 'invscaling', 'adaptive']
        if learning_rate_v not in list_learn:
            raise ValueError(f"Selected learning_rate_v is not compatible. "
                             f"Please try one of the following options: 'constant', 'invscaling', 'adaptive'")

        self.param_frame = NNParameter(activation_v, solver_v, alpha_v, batch_size_v,
                 learning_rate_v, learning_rate_init_v, power_t_v, max_iter_v,
                 shuffle_v, tol_v, verbose_v, warm_start_v, momentum_v,
                 nesterovs_momentum_v, early_stopping_v, validation_fraction_v,
                 beta_1_v, beta_2_v, n_iter_no_change_v, max_fun_v)
        
        self.save_model = save_model
        self.num_teste = num_teste
        self.por_trei = por_trei
        self.generate_preview_nn()

    def nearest_neighbors(self, n_neighbors_v=5, algorithm_v='auto', leaf_size_v=30,
                          p_v=2, n_jobs_v='5', por_trei=70, num_teste=5, save_model=False, load_params='None'):
        
        if load_params != 'None':
            params = self._load_json_params(load_params)
            if params:
                self.param_frame = NNeighParameter(
                    params["n_neighbors"],
                    'auto',
                    30,
                    params["p_value"],
                    '-1'
                )
                self.save_model = save_model
                self.num_teste = num_teste
                self.por_trei = por_trei
                self.generate_preview_kn()
                return
        
        list_alg = ['auto', 'ball_tree', 'kd_tree', 'brute']
        if algorithm_v not in list_alg:
            raise ValueError(f"Selected algorithm_v is not compatible. "
                             f"Please try one of the following options: 'auto', 'ball_tree', 'kd_tree', 'brute'")
        self.param_frame = NNeighParameter(n_neighbors_v, algorithm_v, leaf_size_v, p_v, n_jobs_v)

        self.save_model = save_model
        self.num_teste = num_teste
        self.por_trei = por_trei
        self.generate_preview_kn()

    def support_vector(self, kernel_v='rbf', degree_v=3, gamma_v='scale', coef0_v='0.0',
                       tol_v='0.001', c_v='1.0', epsilon_v='0.1', shrinking_v=True,
                       cache_size_v='200', verbose_v=False, maxiter_v=-1, por_trei=70,
                       num_teste=5, save_model=False, load_params='None'):
        
        if load_params != 'None':
            params = self._load_json_params(load_params)
            if params:
                self.kernel_v = 'rbf'
                self.gamma_v = params["gamma"]
                self.c_v = params["c_param"]
                self.epsilon_v = params["epsilon"]
                self.por_trei = por_trei
                self.num_teste = num_teste
                self.save_model = save_model
                self.generate_preview_svm()
                return
        
        list_kernel = ['linear', 'poly', 'rbf', 'sigmoid']
        if kernel_v not in list_kernel:
            raise ValueError(f"Selected kernel_v is not compatible. "
                             f"Please try one of the following options: 'linear', 'poly', 'rbf', 'sigmoid'")
        
        self.kernel_v = kernel_v
        self.degree_v = degree_v
        list_gamma = ['scale', 'auto', 'float']
        if gamma_v not in list_gamma:
            raise ValueError(f"Selected algorithm_v is not compatible. "
                             f"Please try one of the following options: 'auto', 'sacle', 'float'")
        
        self.gamma_v = gamma_v
        self.coef0_v = coef0_v
        self.tol_v = tol_v
        self.c_v = c_v
        self.epsilon_v = epsilon_v
        self.shrinking_v = shrinking_v
        self.cache_size_v = cache_size_v
        self.verbose_v = verbose_v
        self.maxiter_v = maxiter_v
        self.por_trei = por_trei
        self.num_teste = num_teste
        self.save_model = save_model
        self.generate_preview_svm()

    def gaussian_process(self, alpha_gp='0.0000000001', n_restarts_op=0,
                         normalize_y_gp=False, copy_X_train=False, rand_state_gp='None',
                         por_trei=70, num_teste=5, save_model=False, load_params='None'):
        if load_params != 'None':
            params = self._load_json_params(load_params)
            if params:
                self.param_frame = GPParameter(
                    params.get("alpha_gp", alpha_gp),
                    n_restarts_op,
                    normalize_y_gp,
                    copy_X_train,
                    rand_state_gp
                )
                self.por_trei = por_trei
                self.num_teste = num_teste
                self.save_model = save_model
                self.generate_preview_svm()
                return
            
        self.param_frame = GPParameter(alpha_gp, n_restarts_op, normalize_y_gp, copy_X_train, rand_state_gp)
        self.por_trei = por_trei
        self.num_teste = num_teste
        self.save_model = save_model
        self.generate_preview_svm()

    def _get_common_params(self, city, indicator, split, tests, save_model):
        """Prepara os caminhos e parâmetros comuns"""
        base_dir = os.path.abspath(os.path.dirname(sys.argv[0]))

        file_map = {
            "Target city": "target_clean.txt",
            "Neighbor A": "neighborA_clean.txt",
            "Neighbor B": "neighborB_clean.txt",
            "Neighbor C": "neighborC.txt"
        }

        city_file_name = file_map.get(city, "target_clean.txt")
        city_path = os.path.join(base_dir, city_file_name)

        if not os.path.exists(city_path):
            print(f"Erro: Arquivo não encontrado em {city_path}")
            return None
        
        return (
            city_path,
            get_indicator_code(indicator),
            split,
            tests,
            save_model
        )

    def optuna_optimization(self, n_trials, model_name, city, indicator, split, tests, 
                            save_model=True, gp_kernel_type='RBF'):
        common_params = self._get_common_params(city, indicator, split, tests, save_model)

        if common_params is None: return
        
        objective_mapping = {
            'Decision Trees': self.objective_dt,
            'Bagged Trees': self.objective_bt,
            'Neural network': self.objective_nn,
            'Nearest Neighbors': self.objective_kn,
            'Support Vector': self.objective_svm,
            'Gaussian Process': self.objective_gp,
        }

        base_objective = objective_mapping.get(model_name)
        if base_objective is None:
            print(f"Modelo {model_name} não suportado.")
            return
        
        print(f"--- Iniciando Otimização para {model_name} ({n_trials} trials) ---")
        
        optuna.logging.set_verbosity(optuna.logging.INFO)
        study = optuna.create_study(direction='maximize')

        if model_name == 'Gaussian Process':
            objective_func = functools.partial(base_objective, common_params=common_params, kernel_type=gp_kernel_type)
        else:
            objective_func = functools(base_objective, common_params=common_params)

        try:
            study.optimize(objective_func, n_trials=n_trials)
        except Exception as e:
            print(f"Erro durante a otimização: {e}")
            return
        
        self.last_best_params = study.best_params
        best_value = study.best_value
        best_params = study.best_params

        print("\n" + "="*40)
        print(f"Otimização Finalizada!")
        print(f"Melhor Score (R2/Metrics): {best_value}")
        print(f"Melhores Hiperparâmetros: {best_params}")
        print("="*40 + "\n")

        print(f"Salvando JSON de hiperparâmetros...")
        self.save_hyperparameters(model_name)

        print(f"Gerando modelo final para {model_name} com os melhores parâmetros...")
        self._train_final_model(model_name, best_params, common_params, gp_kernel_type)

    def _train_final_model(self, model_name, best_params, common_params, gp_kernel_type):
        """
        Treina o modelo com os melhores parâmetros encontrados
        """
        city_path, indicator, split_percentage, n_tests, save_model = common_params
        trainer = Training()

        try:
            if model_name == 'Decision Trees':
                trainer.decision_tree(
                    city_path, indicator, split_percentage,
                    criterion=best_params['criterion'],
                    splitter=best_params['splitter'],
                    max_depth=best_params['max_depth'],
                    min_samples_split=best_params['min_samples_split'],
                    min_samples_leaf=best_params['min_samples_leaf'],
                    max_features=best_params['max_features'],
                    min_impurity_decrease=best_params['min_impurity_decrease'],
                    ccp_alpha=best_params['ccp_alpha'],
                    n_tests=n_tests,
                    save_model=save_model
                )

            elif model_name == 'Bagged Trees':
                trainer.bagging_trees(
                    city_path, indicator, split_percentage,
                    criterion=best_params['criterion'],
                    splitter='best',
                    max_depth=best_params['max_depth'],
                    min_samples_leaf=best_params['min_samples_leaf'],
                    max_features=best_params['max_features'],
                    min_samples_split=best_params['min_samples_split'],
                    min_impurity_decrease=best_params['min_impurity_decrease'],
                    ccp_alpha=best_params['ccp_alpha'],
                    n_estimators=best_params['n_estimators_bag'],
                    n_tests=n_tests,
                    save_model=save_model
                )

            elif model_name == 'Neural network':
                momentum = best_params.get('momentum', 0.9)
                batch_size = int(best_params['batch_size']) if best_params['batch_size'] != 'auto' else 'auto'
                
                trainer.neural_network(
                    city_path, indicator, split_percentage, n_tests,
                    activation=best_params['activation'],
                    solver=best_params['solver'],
                    alpha=best_params['alpha'],
                    batch_size=batch_size,
                    learning_rate=best_params['learning_rate'],
                    learning_rate_init=best_params['learning_rate_init'],
                    max_iter=best_params['max_iter'],
                    momentum=momentum,
                    save_model=save_model
                )

            elif model_name == 'Support Vector':
                trainer.support_vector_regression(
                    city_path, indicator, split_percentage, n_tests,
                    kernel='rbf',
                    degree=3,
                    gamma=best_params['gamma'],
                    C=best_params['c_param'],
                    epsilon=best_params['epsilon'],
                    save_model=save_model
                )
            
            elif model_name == 'Nearest Neighbors':
                trainer.KNeighbors(
                    city_path, indicator, split_percentage, n_tests,
                    n_neighbors=best_params['n_neighbors'],
                    p=best_params['p_value'],
                    algorithm='auto',
                    leaf_size=30,
                    n_jobs=-1,
                    save_model=save_model
                )

            elif model_name == 'Gaussian Process':
                trainer.gaussian_process_regression(
                    city_path, indicator, split_percentage, n_tests,
                    kernel=gp_kernel_type,
                    len_scale=best_params.get('length_scale', 1.0),
                    nu=best_params.get('nu', 1.5),
                    sigma_0=best_params.get('sigma_0', 1.0),
                    alpha_rq=best_params.get('alpha_rq', 1.0),
                    alpha_gp=best_params.get('alpha_gp', 1e-10),
                    n_restarts_optimizer=0,
                    save_model=save_model
                )
            
            print(f"Sucesso: Modelo {model_name} salvo corretamente.")
            
        except Exception as e:
            print(f"Erro ao salvar o modelo final: {e}")

    def save_hyperparameters(self, model_name):
        """
        Salva os parâmetros otimizados automaticamente em um arquivo JSON
        no diretório de execução, sem abrir janelas.
        """
        if not self.last_best_params:
            print("Aviso: Nenhum parâmetro encontrado para salvar.")
            return

        safe_name = model_name.replace(" ", "_")
        file_name = f"params_{safe_name}.json"
        
        file_path = os.path.join(os.getcwd(), file_name)

        try:
            data_to_save = {
                "model": model_name,
                "params": self.last_best_params
            }
            
            with open(file_path, 'w', encoding='utf-8') as f:
                json.dump(data_to_save, f, indent=4)
            
            print(f"Sucesso: Hiperparâmetros salvos em: {file_path}")
        except Exception as e:
            print(f"Erro ao salvar arquivo JSON: {str(e)}")

    def objective_dt(self, trial, common_params):
        """Função objetivo para Decision Trees"""
        city_path, indicator, split_percentage, n_tests, save_model_flag = common_params
        
        criterion = trial.suggest_categorical('criterion', ['squared_error', 'absolute_error'])
        splitter = trial.suggest_categorical('splitter', ['best', 'random'])
        max_depth = trial.suggest_int('max_depth', 5, 50)
        min_samples_split = trial.suggest_int('min_samples_split', 2, 20)
        min_samples_leaf = trial.suggest_int('min_samples_leaf', 1, 20)
        max_features = trial.suggest_float('max_features', 0.5, 1.0)
        
        min_weight_fraction_leaf = 0.0 
        
        min_impurity_decrease = trial.suggest_float('min_impurity_decrease', 0.0, 0.01)
        
        ccp_alpha = trial.suggest_float('ccp_alpha', 1e-5, 0.005, log=True)

        prev = Training()
        score, *_ = prev.decision_tree(
            city_path, indicator, split_percentage, criterion, splitter, max_depth,
            min_samples_leaf, max_features, 
            max_leaf_nodes=None,
            n_tests=n_tests, 
            min_samples_split=min_samples_split,
            min_weight_fraction_leaf=min_weight_fraction_leaf, 
            min_impurity_decrease=min_impurity_decrease, 
            ccp_alpha=ccp_alpha,
            save_model=False
        )
        return score
    
    def objective_bt(self, trial, common_params):
        """Função objetivo para Bagged Trees (VERSÃO CORRIGIDA)."""
        city_path, indicator, split_percentage, n_tests, save_model_flag = common_params

        n_estimators = trial.suggest_int('n_estimators_bag', 50, 200, log=True) 

        criterion = trial.suggest_categorical('criterion', ['squared_error', 'absolute_error'])
        max_depth = trial.suggest_int('max_depth', 5, 50)
        min_samples_split = trial.suggest_int('min_samples_split', 2, 20)
        min_samples_leaf = trial.suggest_int('min_samples_leaf', 1, 20)
        max_features = trial.suggest_float('max_features', 0.5, 1.0)
        
        min_weight_fraction_leaf = 0.0
        min_impurity_decrease = trial.suggest_float('min_impurity_decrease', 0.0, 0.01)
        ccp_alpha = trial.suggest_float('ccp_alpha', 1e-5, 0.005, log=True)
        
        prev = Training()
        score, *_ = prev.bagging_trees(
             city_path, indicator, split_percentage, criterion, 
             splitter='best',
             max_depth=max_depth,
             min_samples_leaf=min_samples_leaf, 
             max_features=max_features, 
             max_leaf_nodes=None, 
             n_tests=n_tests, min_samples_split=min_samples_split,
             min_weight_fraction_leaf=min_weight_fraction_leaf, 
             min_impurity_decrease=min_impurity_decrease, 
             ccp_alpha=ccp_alpha, 
             save_model=False,
             n_estimators=n_estimators
        )
        return score
    
    def objective_nn(self, trial, common_params):
        """Função objetivo para Neural Network."""
        city_path, indicator, split_percentage, n_tests, save_model_flag = common_params

        activation = trial.suggest_categorical('activation', ['relu', 'tanh', 'logistic'])
        solver = trial.suggest_categorical('solver', ['adam', 'lbfgs', 'sgd'])
        alpha = trial.suggest_float('alpha', 1e-5, 1e-1, log=True)
        batch_size_str = trial.suggest_categorical('batch_size', ['auto', '50', '100', '200'])
        batch_size = int(batch_size_str) if batch_size_str != 'auto' else 'auto'
        learning_rate = trial.suggest_categorical('learning_rate', ['constant', 'invscaling', 'adaptive'])
        learning_rate_init = trial.suggest_float('learning_rate_init', 1e-4, 1e-2, log=True)
        max_iter = trial.suggest_int('max_iter', 200, 1000)
        momentum = 0.9
        
        if solver == 'sgd':
            momentum = trial.suggest_float('momentum', 0.1, 0.99)
        
        prev = Training()
        score, *_ = prev.neural_network(
             city_path, indicator, split_percentage, n_tests, activation, solver, alpha, batch_size,
             learning_rate, learning_rate_init, power_t=0.5, max_iter=max_iter, shuffle=True, 
             tol=1e-4, verbose=False, warm_start=False, momentum=momentum, 
             nesterovs_momentum=True, early_stopping=False, validation_fraction=0.1,
             beta_1=0.9, beta_2=0.999, n_iter_no_change=10, max_fun=15000, 
             save_model=False
        )
        return score

    def objective_svm(self, trial, common_params):
        """Função objetivo para Support Vector."""
        city_path, indicator, split_percentage, n_tests, save_model_flag = common_params

        kernel = 'rbf'
        c_param = trial.suggest_float('c_param', 1e-2, 1e2, log=True)
        epsilon = trial.suggest_float('epsilon', 1e-3, 1e-1, log=True)
        
        degree = 3
        gamma = trial.suggest_float('gamma', 1e-4, 1.0, log=True)
        
        prev = Training()
        score, *_ = prev.support_vector_regression(
             city_path, indicator, split_percentage, n_tests, kernel, degree, gamma, 
             coef0=0.0, tol=1e-3, C=c_param, epsilon=epsilon, 
             shrinking=True, cache_size=200, verbose=False, max_iter=-1, 
             save_model=False
        )
        return score

    def objective_gp(self, trial, common_params, kernel_type='RBF'):
        """
        Função objetivo do gaussian process.
        """
        city_path, indicator, split_percentage, n_tests, _ = common_params
        
        length_scale = 1.0
        nu = 1.5
        alpha_rq = 1.0
        sigma_0 = 1.0

        if kernel_type in ['RBF', 'Matern', 'RationalQuadratic']:
            length_scale = trial.suggest_float('length_scale', 1e-2, 1e2, log=True)
        
        if kernel_type == 'Matern':
            nu = trial.suggest_categorical('nu', [0.5, 1.5, 2.5])
            
        if kernel_type == 'RationalQuadratic':
            alpha_rq = trial.suggest_float('alpha_rq', 1e-2, 1e2, log=True)
            
        if kernel_type == 'DotProduct':
            sigma_0 = trial.suggest_float('sigma_0', 1e-2, 10.0, log=True)
        
        alpha_gp = trial.suggest_float('alpha_gp', 1e-10, 1e-1, log=True)

        prev = Training()
        score, *_ = prev.gaussian_process_regression(
             city_path, indicator, split_percentage, n_tests, kernel_type, 
             len_scale=length_scale, nu=nu, sigma_0=sigma_0, alpha_rq=alpha_rq,
             alpha_gp=alpha_gp, n_restarts_optimizer=0, normalize_y_gp=False, save_model=False
        )
        return score


    def objective_kn(self, trial, common_params):
        """Função objetivo para Nearest Neighbors (ESPAÇO REDUZIDO)."""
        city_path, indicator, split_percentage, n_tests, save_model_flag = common_params
        
        n_neighbors = trial.suggest_int('n_neighbors', 3, 30)
        
        p_value = trial.suggest_int('p_value', 1, 2) # 1=Manhattan, 2=Euclidean

        algorithm = 'auto'
        leaf_size = 30   

        prev = Training()
        score, *_ = prev.KNeighbors(
             city_path, indicator, split_percentage, n_tests, 
             n_neighbors, 
             algorithm,  
             leaf_size,   
             p_value,     
             n_jobs=-1,
             save_model=False
        )
        return score



import tkinter as tk
from tkinter import ttk, messagebox as msg, filedialog
from tkinter.scrolledtext import ScrolledText
import typing
import os
import sys
import optuna
import functools
import logging
from ...tooltip import CreateToolTip
import threading
import json

from ....training.training import Training
from ....utils.indicator import get_indicator_code 
from .optuna_popup import OptunaPopup

from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk

if typing.TYPE_CHECKING:
    from ...start_page import StartPage 
    from .gui_helpers import DTParameterFrame, BaggingTParameterFrame, NNParameterFrame, NNeighParameterFrame, SVMParameterFrame, GPParameterFrame

class TextHandler(logging.Handler):
    """
    Classe customizada para redirecionar logs para um widget ScrolledText do Tkinter.
    """
    def __init__(self, text_widget):
        super().__init__()
        self.text_widget = text_widget

    def emit(self, record):
        msg = self.format(record)
        def append():
            self.text_widget.configure(state='normal')
            self.text_widget.insert(tk.END, msg + '\n')
            self.text_widget.configure(state='disabled')
            self.text_widget.see(tk.END)
        
        self.text_widget.after(0, append)

class MachineLearningPage(ttk.Frame):
    """Tela para treinamento e visualização de modelos de Machine Learning."""
    def __init__(self, parent, controller):
        ttk.Frame.__init__(self, parent)
        self.controller = controller
        i18n = controller.i18n

        # --- Variáveis de Controle ---
        self.split_var = tk.StringVar(value="70")
        self.tests_var = tk.StringVar(value="10")
        self.save_model_var = tk.BooleanVar(value=False)
        self.optuna_trials_var = tk.StringVar(value="10")
        self.current_param_frame = None

        # --- Frame Superior ---
        top_frame = ttk.Frame(self)
        top_frame.pack(fill=tk.X, padx=10, pady=10)
        self.back_button = ttk.Button(top_frame, text="", command=self.go_to_previous_page)
        self.back_button.pack(side=tk.LEFT)
        self.page_title = ttk.Label(top_frame, text="", font=("Verdana", 16, "bold"))
        self.page_title.pack(side=tk.LEFT, expand=True)

        # --- Layout Principal ---
        main_container = ttk.Frame(self)
        main_container.pack(fill=tk.BOTH, expand=True, padx=10, pady=10)
        main_container.grid_columnconfigure(0, weight=1, minsize=300)
        main_container.grid_columnconfigure(1, weight=3)
        main_container.grid_rowconfigure(0, weight=1)

        # --- PAINEL ESQUERDO ---
        left_panel_base = ttk.Frame(main_container)
        left_panel_base.grid(row=0, column=0, sticky="nsew", padx=(0, 10))
        left_panel_base.grid_rowconfigure(0, weight=1)
        left_panel_base.grid_columnconfigure(0, weight=1)

        self.canvas_left = tk.Canvas(left_panel_base)
        self.canvas_left.grid(row=0, column=0, sticky="nsew")

        scrollbar_left = ttk.Scrollbar(left_panel_base, orient="vertical", command=self.canvas_left.yview)
        scrollbar_left.grid(row=0, column=1, sticky="ns")
        self.canvas_left.configure(yscrollcommand=scrollbar_left.set)

        self.scrollable_frame_left = ttk.Frame(self.canvas_left)

        self.canvas_left.create_window((0, 0), window=self.scrollable_frame_left, anchor="nw")

        self.scrollable_frame_left.bind(
            "<Configure>",
            lambda e: self.canvas_left.configure(
                scrollregion=self.canvas_left.bbox("all")
            )
        )

        self.canvas_left.bind_all("<MouseWheel>", self._on_mousewheel) # Linux/Windows
        self.canvas_left.bind_all("<Button-4>", self._on_mousewheel) # Linux scroll up
        self.canvas_left.bind_all("<Button-5>", self._on_mousewheel) # Linux scroll down

        # --- Configurações Gerais ---
        general_frame = ttk.LabelFrame(self.scrollable_frame_left, text="")
        general_frame.pack(fill=tk.X, pady=(0, 10), padx=5)
        self.general_frame_label = general_frame

        self.data_label = ttk.Label(general_frame, text="")
        self.data_label.pack(anchor="w", padx=5)
        self.data_combo = ttk.Combobox(general_frame, state="readonly", values=["Target", "Neighbor A", "Neighbor B", "Neighbor C"])
        self.data_combo.pack(fill=tk.X, padx=5, pady=(0, 5))
        self.data_combo_hint = CreateToolTip(self.data_combo, text=i18n.get('data_MaL_hint'))

        self.indicator_label = ttk.Label(general_frame, text="")
        self.indicator_label.pack(anchor="w", padx=5)
        self.indicator_combo = ttk.Combobox(general_frame, state="readonly", values=["Precipitation", "Maximum temperature", "Minimum temperature"])
        self.indicator_combo.pack(fill=tk.X, padx=5, pady=(0, 10))
        self.indicator_combo_hint = CreateToolTip(self.indicator_combo, text=i18n.get('param_MaL_hint'))

        self.training_percentage_label = ttk.Label(general_frame, text="")
        self.training_percentage_label.pack(anchor="w", padx=5)
        self.split_entry = ttk.Entry(general_frame, textvariable=self.split_var)
        self.split_entry.pack(fill=tk.X, padx=5, pady=(0, 5))
        self.split_entry_hint = CreateToolTip(self.split_entry, text=i18n.get('train_percent_MaL_hint'))

        self.number_tests_label = ttk.Label(general_frame, text="")
        self.number_tests_label.pack(anchor="w", padx=5)
        self.tests_entry = ttk.Entry(general_frame, textvariable=self.tests_var)
        self.tests_entry.pack(fill=tk.X, padx=5, pady=(0, 10))
        self.tests_entry_hint = CreateToolTip(self.tests_entry, text=i18n.get('number_tests_MaL_hint'))

        self.save_model_check = ttk.Checkbutton(general_frame, text="", variable=self.save_model_var)
        self.save_model_check.pack(anchor="w", padx=5, pady=(0, 5))

        # --- Seleção do Modelo ---
        model_frame = ttk.LabelFrame(self.scrollable_frame_left, text="")
        model_frame.pack(fill=tk.X, pady=(0, 10), padx=5)
        self.model_frame_label = model_frame

        self.model_label = ttk.Label(model_frame, text="")
        self.model_label.pack(anchor="w", padx=5)
        model_list = ['Decision Trees', 'Bagged Trees', 'Neural network', 'Nearest Neighbors', 'Support Vector', 'Gaussian Process']
        self.ml_combo = ttk.Combobox(model_frame, state="readonly", values=model_list)
        self.ml_combo.pack(fill=tk.X, padx=5, pady=(0, 5))
        self.ml_combo_hint = CreateToolTip(self.ml_combo, text=i18n.get('model_MaL_hint'))
        self.ml_combo.bind("<<ComboboxSelected>>", self.generate_param_ui)

        # --- Parâmetros do Modelo ---
        self.param_container = ttk.LabelFrame(self.scrollable_frame_left, text="")
        self.param_container.pack(fill=tk.X, pady=(0, 10), padx=5)

        """
        self.optuna_trials_label = ttk.Label(self.scrollable_frame_left, text="")
        self.optuna_trials_label.pack(anchor="w", padx=5, pady=(10, 0))

        self.optuna_trials_entry = ttk.Entry(self.scrollable_frame_left, textvariable=self.optuna_trials_var)
        self.optuna_trials_entry.pack(fill=tk.X, padx=5, pady=(0, 5))
        self.optuna_trials_hint = CreateToolTip(self.optuna_trials_entry, text=i18n.get("optuna_trials_MaL_hint"))
        """
        self.optuna_btn = ttk.Button(self.scrollable_frame_left, 
                                     text="", 
                                     command=self.run_optuna_optimization)
        self.optuna_btn.pack(fill=tk.X, pady=(10, 0), padx=5)
        self.optuna_btn_hint = CreateToolTip(self.optuna_btn, text=i18n.get('optuna_MaL_hint'))

        manage_frame = ttk.Frame(self.scrollable_frame_left)
        manage_frame.pack(fill=tk.X, pady=5, padx=5)

        self.btn_save_params = ttk.Button(manage_frame, text="Salvar Params", command=self.save_hyperparameters)
        self.btn_save_params.pack(side=tk.LEFT, fill=tk.X, expand=True, padx=(0, 2))
        
        self.btn_save_params.config(state="disabled") 

        self.btn_load_params = ttk.Button(manage_frame, text="Carregar Params", command=self.load_hyperparameters)
        self.btn_load_params.pack(side=tk.RIGHT, fill=tk.X, expand=True, padx=(2, 0))

        self.last_best_params = None

        # --- Executar ---
        self.run_btn = ttk.Button(self.scrollable_frame_left, text="", command=self.run_model_preview)
        self.run_btn.pack(fill=tk.X, pady=10, padx=5)
        self.run_btn_hint = CreateToolTip(self.run_btn  , text=i18n.get('generate_prev_MaL_hint'))

        # --- PAINEL DIREITO ---
        self.right_panel = ttk.Frame(main_container, relief="solid", borderwidth=1)
        self.right_panel.grid(row=0, column=1, sticky="nsew")

        self.update_texts() # Chama para definir os textos iniciais

    def _on_mousewheel(self, event):
        """Função para habilitar scroll com a roda do mouse."""
        # Determina a direção e a quantidade do scroll
        if event.num == 4 or event.delta > 0: # Linux scroll up ou Windows/macOS scroll up
            self.canvas_left.yview_scroll(-1, "units")
        elif event.num == 5 or event.delta < 0: # Linux scroll down ou Windows/macOS scroll down
            self.canvas_left.yview_scroll(1, "units")

    def go_to_previous_page(self):
        """Importa e navega para a página anterior."""
        from ..imputation_techniques_page import ImputationTechniquesPage
        self.controller.show_frame(ImputationTechniquesPage)

    def _clear_panel(self, panel):
        """Remove todos os widgets de um painel."""
        for widget in panel.winfo_children():
            widget.destroy()

    def update_texts(self):
        """Atualiza os textos estáticos da tela."""
        i18n = self.controller.i18n
        self.back_button.config(text=i18n.get('back_btn'))
        self.page_title.config(text=i18n.get('ml_page_title'))
        self.general_frame_label.config(text=i18n.get('general_settings_label'))
        self.save_model_check.config(text=i18n.get('save_model_label'))
        self.model_frame_label.config(text=i18n.get('model_selection_label'))
        self.param_container.config(text=i18n.get('model_parameters_label'))
        self.run_btn.config(text=i18n.get('run_prediction_btn'))
        self.data_label.config(text=i18n.get('data_label'))
        self.indicator_label.config(text=i18n.get('parameter_label'))
        self.model_label.config(text=i18n.get('model_label'))
        self.number_tests_label.config(text=i18n.get('number_tests_label'))
        self.training_percentage_label.config(text=i18n.get('training_percentage_label'))
        self.optuna_btn.config(text=i18n.get('param_optimizer_btn'))
        #self.optuna_trials_label.config(text="Número de tentativas")
        self.data_combo_hint.text = i18n.get('data_MaL_hint')
        self.indicator_combo_hint.text = i18n.get('param_MaL_hint')
        self.split_entry_hint.text = i18n.get('train_percent_MaL_hint')
        self.tests_entry_hint.text = i18n.get('number_tests_MaL_hint')
        self.ml_combo_hint.text = i18n.get('model_MaL_hint')
        #self.optuna_trials_hint.text = i18n.get('optuna_trials_MaL_hint')
        self.optuna_btn_hint.text = i18n.get('optuna_MaL_hint')
        self.optuna_btn_hint.text = i18n.get('generate_prev_MaL_hint')

    def _setup_log_view(self):
        """Prepara o painel direito para exibir logs."""
        self._clear_panel(self.right_panel)
        
        lbl = ttk.Label(self.right_panel, text="Progresso da Otimização (Optuna):", font=("Arial", 10, "bold"))
        lbl.pack(anchor="w", padx=10, pady=(10, 5))
        
        log_text = ScrolledText(self.right_panel, state='disabled', height=20)
        log_text.pack(fill=tk.BOTH, expand=True, padx=10, pady=10)
        
        return log_text

    def run_optuna_optimization(self):
        model_name = self.ml_combo.get()
        
        if not model_name:
            self.controller.show_translated_message(
                 msg_type='warning',
                 title_key='warning_title',
                 message_key='select_ml_model_msg'
             )
            return
        
        if self._get_common_params() is None:
            return

        OptunaPopup(self, self.controller, callback=self._execute_optuna_optimization)


    def _execute_optuna_optimization(self, n_trials):
        model_name = self.ml_combo.get()
        common_params = self._get_common_params()
        
        objective_mapping = {
            'Decision Trees': self.objective_dt,
            'Bagged Trees': self.objective_bt,
            'Neural network': self.objective_nn,
            'Nearest Neighbors': self.objective_kn,
            'Support Vector': self.objective_svm,
            'Gaussian Process': self.objective_gp,
        }

        base_objective = objective_mapping.get(model_name)
        if base_objective is None:
            return

        log_widget = self._setup_log_view()
        
        optuna.logging.set_verbosity(optuna.logging.INFO)
        optuna_logger = optuna.logging.get_logger("optuna")
        
        handler = TextHandler(log_widget)
        handler.setFormatter(logging.Formatter('%(message)s'))
        optuna_logger.addHandler(handler)

        def run_optimization_thread():
            try:
                objective_func = functools.partial(base_objective, common_params=common_params)
                
                study = optuna.create_study(direction='maximize')
                study.optimize(objective_func, n_trials=n_trials)

                best_params = study.best_params
                best_value = study.best_value

                self.controller.after(0, lambda: self._on_optimization_finished(best_params, best_value, optuna_logger, handler))

            except Exception as e:
                self.controller.after(0, lambda: self.controller.show_translated_message(
                    msg_type='error', 
                    title_key='optimization_error_title', 
                    message_key='optimization_error_msg', 
                    error=str(e)
                ))
                self.controller.after(0, lambda: optuna_logger.removeHandler(handler))

        threading.Thread(target=run_optimization_thread, daemon=True).start()

    def _on_optimization_finished(self, best_params, best_value, logger, handler):
        """Chamado quando a thread termina com sucesso."""
        
        logger.removeHandler(handler)

        self.last_best_params = best_params
        self.btn_save_params.config(state="normal")
        
        self.update_ui_with_params(best_params)

        self.controller.show_translated_message(
            msg_type='info',
            title_key='optimization_complete_title',
            message_key='optimization_complete_msg',
            best_score=f"{best_value:.4f}",
            params=str(best_params)
        )
    def update_ui_with_params(self, best_params):
        """
        Preenche os campos da UI (Entry, Combobox) com os 
        parâmetros do dicionário 'best_params' do Optuna.
        """

        if not self.current_param_frame:
            return
        
        param_map = {
            # Decision Trees / Bagging
            'criterion': 'criterion_v',
            'splitter': 'splitter_v',
            'max_depth': 'maxd_v',
            'min_samples_split': 'minsam_s_v',
            'min_samples_leaf': 'minsam_l_v',
            'min_weight_fraction_leaf': 'minweifra_l_v',
            'max_features': 'maxfeat_v',
            'max_leaf_nodes': 'maxleaf_n',
            'min_impurity_decrease': 'minimp_dec',
            'ccp_alpha': 'ccp_alp_v',
            # Bagging (nomeado diferente para evitar colisão)
            'n_estimators_bag': 'n_estimators', 
            # Neural Network
            'activation': 'activation_v',
            'solver': 'solver_v',
            'alpha': 'alpha_v',
            'batch_size': 'batch_size_v',
            'learning_rate': 'learning_rate_v',
            'learning_rate_init': 'learning_rate_init_v',
            'max_iter': 'max_iter_v',
            'momentum': 'momentum_v',
            # Nearest Neighbors
            'n_neighbors': 'n_neighbors_v',
            'algorithm': 'algorithm_v',
            'leaf_size': 'leaf_size_v',
            'p_value': 'p_v',
            # SVM
            'kernel': 'kernel_v',
            'degree': 'degree_v',
            'gamma': 'gamma_v',
            'c_param': 'c_v',
            'epsilon': 'epsilon_v',
            # Gaussian Process
            'alpha_gp': 'alpha_gp',
            'length_scale': 'length_scale',
            'nu': 'nu',
            'sigma_0': 'sigma_0',
            'alpha_rq': 'alpha_rq',
        }

        for param_name, widget_var_name in param_map.items():
            if param_name in best_params:
                if hasattr(self.current_param_frame, widget_var_name):
                    widget_var = getattr(self.current_param_frame, widget_var_name)
                    widget_var.set(str(best_params[param_name]))
                else:
                    print(f"Aviso: Parâmetro '{param_name}' encontrado, mas UI var '{widget_var_name}' não existe no frame atual.")

    def save_hyperparameters(self):
        """Salva os últimos parâmetros otimizados em um arquivo JSON."""
        if not self.last_best_params:
            return

        model_name = self.ml_combo.get().replace(" ", "_")
        
        file_path = filedialog.asksaveasfilename(
            defaultextension=".json",
            filetypes=[("JSON files", "*.json"), ("All files", "*.*")],
            initialfile=f"params_{model_name}.json",
            title="Salvar Hiperparâmetros"
        )

        if file_path:
            try:
                data_to_save = {
                    "model": self.ml_combo.get(),
                    "params": self.last_best_params
                }
                
                with open(file_path, 'w') as f:
                    json.dump(data_to_save, f, indent=4)
                
                msg.showinfo("Sucesso", "Parâmetros salvos com sucesso!")
            except Exception as e:
                msg.showerror("Erro", f"Falha ao salvar arquivo: {str(e)}")

    def load_hyperparameters(self):
        """Carrega parâmetros de um arquivo JSON e preenche a UI."""
        file_path = filedialog.askopenfilename(
            filetypes=[("JSON files", "*.json"), ("All files", "*.*")],
            title="Carregar Hiperparâmetros"
        )

        if file_path:
            try:
                with open(file_path, 'r') as f:
                    data = json.load(f)

                if "params" in data:
                    params_to_load = data["params"]
                    model_saved = data.get("model", "Desconhecido")
                    
                    current_model = self.ml_combo.get()
                    if model_saved != "Desconhecido" and model_saved != current_model:
                        confirm = msg.askyesno(
                            "Modelo Diferente", 
                            f"Estes parâmetros foram salvos para '{model_saved}', "
                            f"mas o modelo atual é '{current_model}'.\n\nDeseja carregar mesmo assim?"
                        )
                        if not confirm:
                            return

                else:
                    params_to_load = data

                self.update_ui_with_params(params_to_load)
                
                self.last_best_params = params_to_load
                self.btn_save_params.config(state="normal")

                msg.showinfo("Sucesso", "Parâmetros carregados na interface!")

            except json.JSONDecodeError:
                msg.showerror("Erro", "O arquivo selecionado não é um JSON válido.")
            except Exception as e:
                msg.showerror("Erro", f"Falha ao ler arquivo: {str(e)}")

    def objective_dt(self, trial, common_params):
        """Função objetivo para Decision Trees"""
        city_path, indicator, split_percentage, n_tests, save_model_flag = common_params
        
        criterion = trial.suggest_categorical('criterion', ['squared_error', 'absolute_error'])
        splitter = trial.suggest_categorical('splitter', ['best', 'random'])
        max_depth = trial.suggest_int('max_depth', 5, 50)
        min_samples_split = trial.suggest_int('min_samples_split', 2, 20)
        min_samples_leaf = trial.suggest_int('min_samples_leaf', 1, 20)
        max_features = trial.suggest_float('max_features', 0.5, 1.0)
        
        min_weight_fraction_leaf = 0.0 
        
        min_impurity_decrease = trial.suggest_float('min_impurity_decrease', 0.0, 0.01)
        
        ccp_alpha = trial.suggest_float('ccp_alpha', 1e-5, 0.005, log=True)

        prev = Training()
        score, *_ = prev.decision_tree(
            city_path, indicator, split_percentage, criterion, splitter, max_depth,
            min_samples_leaf, max_features, 
            max_leaf_nodes=None,
            n_tests=n_tests, 
            min_samples_split=min_samples_split,
            min_weight_fraction_leaf=min_weight_fraction_leaf, 
            min_impurity_decrease=min_impurity_decrease, 
            ccp_alpha=ccp_alpha,
            save_model=False
        )
        
        return score
    
    def objective_bt(self, trial, common_params):
        """Função objetivo para Bagged Trees (VERSÃO CORRIGIDA)."""
        city_path, indicator, split_percentage, n_tests, save_model_flag = common_params

        n_estimators = trial.suggest_int('n_estimators_bag', 50, 200, log=True) 

        criterion = trial.suggest_categorical('criterion', ['squared_error', 'absolute_error'])
        max_depth = trial.suggest_int('max_depth', 5, 50)
        min_samples_split = trial.suggest_int('min_samples_split', 2, 20)
        min_samples_leaf = trial.suggest_int('min_samples_leaf', 1, 20)
        max_features = trial.suggest_float('max_features', 0.5, 1.0)
        
        min_weight_fraction_leaf = 0.0
        min_impurity_decrease = trial.suggest_float('min_impurity_decrease', 0.0, 0.01)
        ccp_alpha = trial.suggest_float('ccp_alpha', 1e-5, 0.005, log=True)
        
        prev = Training()
        score, *_ = prev.bagging_trees(
             city_path, indicator, split_percentage, criterion, 
             splitter='best',
             max_depth=max_depth,
             min_samples_leaf=min_samples_leaf, 
             max_features=max_features, 
             max_leaf_nodes=None, 
             n_tests=n_tests, min_samples_split=min_samples_split,
             min_weight_fraction_leaf=min_weight_fraction_leaf, 
             min_impurity_decrease=min_impurity_decrease, 
             ccp_alpha=ccp_alpha, 
             save_model=False,
             n_estimators=n_estimators
        )
        return score
    
    def objective_nn(self, trial, common_params):
        """Função objetivo para Neural Network."""
        city_path, indicator, split_percentage, n_tests, save_model_flag = common_params

        activation = trial.suggest_categorical('activation', ['relu', 'tanh', 'logistic'])
        solver = trial.suggest_categorical('solver', ['adam', 'lbfgs', 'sgd'])
        alpha = trial.suggest_float('alpha', 1e-5, 1e-1, log=True)
        batch_size_str = trial.suggest_categorical('batch_size', ['auto', '50', '100', '200'])
        batch_size = int(batch_size_str) if batch_size_str != 'auto' else 'auto'
        learning_rate = trial.suggest_categorical('learning_rate', ['constant', 'invscaling', 'adaptive'])
        learning_rate_init = trial.suggest_float('learning_rate_init', 1e-4, 1e-2, log=True)
        max_iter = trial.suggest_int('max_iter', 200, 1000)
        momentum = 0.9
        
        if solver == 'sgd':
            momentum = trial.suggest_float('momentum', 0.1, 0.99)
        
        prev = Training()
        score, *_ = prev.neural_network(
             city_path, indicator, split_percentage, n_tests, activation, solver, alpha, batch_size,
             learning_rate, learning_rate_init, power_t=0.5, max_iter=max_iter, shuffle=True, 
             tol=1e-4, verbose=False, warm_start=False, momentum=momentum, 
             nesterovs_momentum=True, early_stopping=False, validation_fraction=0.1,
             beta_1=0.9, beta_2=0.999, n_iter_no_change=10, max_fun=15000, 
             save_model=False
        )
        return score

    def objective_svm(self, trial, common_params):
        """Função objetivo para Support Vector."""
        city_path, indicator, split_percentage, n_tests, save_model_flag = common_params

        kernel = 'rbf'
        c_param = trial.suggest_float('c_param', 1e-2, 1e2, log=True)
        epsilon = trial.suggest_float('epsilon', 1e-3, 1e-1, log=True)
        
        degree = 3
        gamma = trial.suggest_float('gamma', 1e-4, 1.0, log=True)
        
        prev = Training()
        score, *_ = prev.support_vector_regression(
             city_path, indicator, split_percentage, n_tests, kernel, degree, gamma, 
             coef0=0.0, tol=1e-3, C=c_param, epsilon=epsilon, 
             shrinking=True, cache_size=200, verbose=False, max_iter=-1, 
             save_model=False
        )
        return score

    def objective_gp(self, trial, common_params):
        """Função objetivo para Gaussian Process."""
        city_path, indicator, split_percentage, n_tests, save_model_flag = common_params
        
        # Pega o tipo de Kernel selecionado na UI
        kernel_type = self.current_param_frame.kernel_type.get() 
        
        # Otimização de Hiperparâmetros baseada no Kernel
        length_scale = 1.0
        nu = 1.5
        alpha_rq = 1.0
        sigma_0 = 1.0

        # RBF, Matern e RationalQuadratic usam length_scale
        if kernel_type in ['RBF', 'Matern', 'RationalQuadratic']:
            length_scale = trial.suggest_float('length_scale', 1e-2, 1e2, log=True)
        
        if kernel_type == 'Matern':
            nu = trial.suggest_categorical('nu', [0.5, 1.5, 2.5])
            
        if kernel_type == 'RationalQuadratic':
            alpha_rq = trial.suggest_float('alpha_rq', 1e-2, 1e2, log=True)
            
        if kernel_type == 'DotProduct':
            sigma_0 = trial.suggest_float('sigma_0', 1e-2, 10.0, log=True)
        
        # Alpha (Noise) global
        alpha_gp = trial.suggest_float('alpha_gp', 1e-10, 1e-1, log=True)

        prev = Training()
        
        score, *_ = prev.gaussian_process_regression(
             city_path, indicator, split_percentage, n_tests, kernel_type, 
             len_scale=length_scale,
             nu=nu, 
             sigma_0=sigma_0, 
             alpha_rq=alpha_rq,
             alpha_gp=alpha_gp, 
             n_restarts_optimizer=int(self.current_param_frame.n_restarts_op.get()), 
             normalize_y_gp=self.current_param_frame.normalize_y_gp.get(), 
             save_model=False
        )
        return score


    def objective_kn(self, trial, common_params):
        """Função objetivo para Nearest Neighbors (ESPAÇO REDUZIDO)."""
        city_path, indicator, split_percentage, n_tests, save_model_flag = common_params
        
        n_neighbors = trial.suggest_int('n_neighbors', 3, 30)
        
        p_value = trial.suggest_int('p_value', 1, 2) # 1=Manhattan, 2=Euclidean

        algorithm = 'auto'
        leaf_size = 30   

        prev = Training()
        score, *_ = prev.KNeighbors(
             city_path, indicator, split_percentage, n_tests, 
             n_neighbors, 
             algorithm,  
             leaf_size,   
             p_value,     
             n_jobs=-1,
             save_model=False
        )
        return score
    
    def generate_param_ui(self, event=None):
        """Preenche o 'param_container' com os widgets corretos."""
        self._clear_panel(self.param_container)
        self.current_param_frame = None

        opcao = self.ml_combo.get()
        if not opcao: return

        try:
            if opcao == 'Decision Trees':
                from .gui_helpers import DTParameterFrame
                self.current_param_frame = DTParameterFrame(self.param_container)
            elif opcao == 'Bagged Trees':
                from .gui_helpers import BaggingTParameterFrame
                self.current_param_frame = BaggingTParameterFrame(self.param_container)
            elif opcao == 'Neural network':
                from .gui_helpers import NNParameterFrame
                self.current_param_frame = NNParameterFrame(self.param_container)
            elif opcao == 'Nearest Neighbors':
                from .gui_helpers import NNeighParameterFrame
                self.current_param_frame = NNeighParameterFrame(self.param_container)
            elif opcao == 'Support Vector':
                from .gui_helpers import SVMParameterFrame
                self.current_param_frame = SVMParameterFrame(self.param_container)
            elif opcao == 'Gaussian Process':
                from .gui_helpers import GPParameterFrame
                self.current_param_frame = GPParameterFrame(self.param_container)

            if self.current_param_frame:
                self.current_param_frame.pack(fill=tk.BOTH, expand=True, padx=5, pady=5)
            else:
                 ttk.Label(self.param_container, text=f"UI de parâmetros para '{opcao}' não implementada.").pack()

        except ImportError as e:
             self.controller.show_translated_message(
                 msg_type='error',
                 title_key='import_error_title',
                 message_key='param_ui_import_error_msg',
                 option=opcao,
                 error=str(e)
             )
        except Exception as e:
            self.controller.show_translated_message(
                 msg_type='error',
                 title_key='ui_generate_error_title',
                 message_key='param_ui_generate_error_msg',
                 option=opcao,
                 error=str(e)
             )
            
    def data_preview(self, score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis):
        """Exibe os resultados textuais e o gráfico de comparação no painel direito."""
        # 1. Limpa o painel direito de qualquer conteúdo anterior
        self._clear_panel(self.right_panel)
        i18n = self.controller.i18n

        # --- Frame para os Resultados Textuais ---
        # Cria um LabelFrame DENTRO do self.right_panel
        results_outer_frame = ttk.LabelFrame(self.right_panel, text=i18n.get('results_preview_label'))
        results_outer_frame.pack(side=tk.TOP, fill=tk.X, padx=10, pady=(10, 5)) # Pack no topo

        # Frame interno para usar .grid() para alinhar os labels
        results_inner_frame = ttk.Frame(results_outer_frame)
        results_inner_frame.pack(padx=10, pady=10)

        # Configura colunas para o grid interno (label, valor_exato, valor_pred)
        results_inner_frame.grid_columnconfigure(0, weight=1)
        results_inner_frame.grid_columnconfigure(1, weight=1)
        results_inner_frame.grid_columnconfigure(2, weight=1)

        # 2. Cria os Labels para as métricas usando ttk e grid
        ttk.Label(results_inner_frame, text=f"{i18n.get('score_label')}: {score:.2f} pts", font='Arial 11 bold').grid(row=0, column=0, sticky="w", columnspan=3)
        ttk.Label(results_inner_frame, text=f"{i18n.get('mean_abs_error_label')}: {round(mean_abs_error, 4)}", font='Arial 11').grid(row=1, column=0, sticky="w", columnspan=3)
        ttk.Label(results_inner_frame, text=f"{i18n.get('mean_rel_error_label')}: {round(mean_rel_error, 4)}", font='Arial 11').grid(row=2, column=0, sticky="w", columnspan=3)

        # Separador visual
        ttk.Separator(results_inner_frame, orient='horizontal').grid(row=3, column=0, columnspan=3, sticky='ew', pady=5)

        # Maior Erro
        ttk.Label(results_inner_frame, text=f"{i18n.get('max_abs_error_label')}: {round(max_abs_error, 4)}", font='Arial 11').grid(row=4, column=0, sticky="w")
        ttk.Label(results_inner_frame, text=f"{i18n.get('exact_value_label')}: {round(exact_max, 4)}", font='Arial 11').grid(row=4, column=1, sticky="w", padx=10)
        ttk.Label(results_inner_frame, text=f"{i18n.get('prediction_label')}: {round(pred_max, 4)}", font='Arial 11').grid(row=4, column=2, sticky="w", padx=10)

        # Menor Erro
        ttk.Label(results_inner_frame, text=f"{i18n.get('min_abs_error_label')}: {round(min_abs_error, 4)}", font='Arial 11').grid(row=5, column=0, sticky="w")
        ttk.Label(results_inner_frame, text=f"{i18n.get('exact_value_label')}: {round(exact_min, 4)}", font='Arial 11').grid(row=5, column=1, sticky="w", padx=10)
        ttk.Label(results_inner_frame, text=f"{i18n.get('prediction_label')}: {round(pred_min, 4)}", font='Arial 11').grid(row=5, column=2, sticky="w", padx=10)

        # --- Criação e Plotagem do Gráfico Matplotlib ---
        figure = Figure(figsize=(10, 6), dpi=100)
        plot_r = figure.add_subplot(111)
        plot_r.plot(x_axis, y_exact, label=i18n.get('exact_plot_label'), color='green')
        plot_r.plot(x_axis, y_pred, label=i18n.get('prediction_plot_label'), color='red')
        plot_r.legend()
        plot_r.grid(True)
        # TODO: Adicionar i18n para os eixos Y e X
        plot_r.set_ylabel("Temperature(°C)")
        plot_r.set_xlabel(i18n.get('comparisons_label'))
        figure.tight_layout() # Ajusta o layout para evitar sobreposições

        # 4. Integração do Matplotlib com Tkinter DENTRO do self.right_panel
        # Cria o Canvas como filho do self.right_panel
        canvas = FigureCanvasTkAgg(figure, master=self.right_panel)
        canvas.draw()

        # Cria a Toolbar como filha do self.right_panel
        toolbar = NavigationToolbar2Tk(canvas, self.right_panel)
        toolbar.update()

        # Coloca a toolbar embaixo e o canvas ocupando o resto
        toolbar.pack(side=tk.BOTTOM, fill=tk.X)
        canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True, padx=10, pady=(0, 10))

    def run_model_preview(self):
        """Função "despachante" que chama a lógica de ML correta."""
        opcao = self.ml_combo.get()
        if not opcao:
            self.controller.show_translated_message(
                 msg_type='warning',
                 title_key='warning_title',
                 message_key='select_ml_model_msg'
             )
            return
        if not self.current_param_frame:
            self.controller.show_translated_message(
                 msg_type='warning',
                 title_key='warning_title',
                 message_key='select_ml_model_again_msg'
             )
            return

        try:
            if opcao == 'Decision Trees': self.generate_preview_dt()
            elif opcao == 'Bagged Trees': self.generate_preview_bt()
            elif opcao == 'Neural network': self.generate_preview_nn()
            elif opcao == 'Nearest Neighbors': self.generate_preview_kn()
            elif opcao == 'Support Vector': self.generate_preview_svm()
            elif opcao == 'Gaussian Process': self.generate_preview_GP()
        except AttributeError as e:
             self.controller.show_translated_message(
                 msg_type='error',
                 title_key='parameter_error_title',
                 message_key='coud_not_read_parameter_msg',
                 option=opcao,
                 error=str(e)
             )
        except Exception as e:
            self.controller.show_translated_message(
                 msg_type='error',
                 title_key='execution_error_title',
                 message_key='model_error_msg',
                 option=opcao,
                 error=str(e)
             )

    def _get_common_params(self):
        city_file_name = self.data_combo.get() 
        base_dir = os.path.abspath(os.path.dirname(sys.argv[0]))
        print("Base dir", base_dir)
        if city_file_name == "Target":
            city_file_name = "target_clean.txt"
        elif city_file_name == "Neighbor A":
            city_file_name = "neighborA_clean.txt"
        elif city_file_name == "Neighbor B":
            city_file_name == "neighborB_clean.txt"
        else:
            city_file_name = "neighborC.txt"

        city_path = os.path.join(base_dir, city_file_name)

        try:
            print(city_path)
            if not os.path.exists(city_path):
                 self.controller.show_translated_message(
                    msg_type='error',
                    title_key='error_title',
                    message_key='file_not_found_msg',
                    city_path=city_path
                )
                 return None
        except Exception as e:
            self.controller.show_translated_message(
                msg_type='error',
                title_key='error_title',
                message_key='path_not_determinated_msg',
                error=str(e)
                )
            return None

        indicator = self.indicator_combo.get()
        indicator = self.indicator_combo.get()
        if not city_file_name or not indicator:
            self.controller.show_translated_message(
                msg_type='error',
                title_key='error_title',
                message_key='file_and_indicator_not_selected_msg'
                )
            return None
        try:
            split = int(self.split_var.get())
            tests = int(self.tests_var.get())
        except ValueError:
             self.controller.show_translated_message(
                msg_type='error',
                title_key='error_title',
                message_key='train_number_tests_msg'
                )
             return None

        try:
            if not os.path.exists(city_path):
                self.controller.show_translated_message(
                msg_type='error',
                title_key='error_title',
                message_key='file_not_found_msg',
                city_path=city_path
                )
                return None
        except Exception as e:
            self.controller.show_translated_message(
                msg_type='error',
                title_key='error_title',
                message_key='path_not_determinated_msg',
                error=str(e)
                )
            return None

        return (
            city_path,
            get_indicator_code(indicator),
            split,
            tests,
            self.save_model_var.get()
        )

    def int_float(self, value_str):
        """Tenta retornar int, senão float."""
        try:
            return int(value_str)
        except ValueError:
            try:
                return float(value_str)
            except ValueError:
                self.controller.show_translated_message(
                msg_type='error',
                title_key='value_error_title',
                message_key='value_error_msg',
                value_str=value_str
                )
                raise 

    def valid_maxf(self, value_str):
        """Retorna o tipo correto para max_features."""
        value_str = value_str.strip() 
        if value_str.isdigit():
            return int(value_str)
        elif value_str in ['sqrt', 'log2']:
             return value_str
        else:
             try:
                 val_float = float(value_str)
                 if val_float.is_integer():
                     return int(val_float)
                 return val_float
             except ValueError:
                 self.controller.show_translated_message(
                    msg_type='error',
                    title_key='value_error_title',
                    message_key='invalid_max_features_msg',
                    value_str=value_str
                    )
                 raise 

    def get_end(self, city_filename):
        """Obtém o caminho completo para um arquivo de cidade processado."""
        try:
            # Precisa encontrar a pasta 'processed_data'
            base_dir = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
            processed_dir = os.path.join(base_dir, "processed_data")
            full_path = os.path.join(processed_dir, city_filename)
            if os.path.exists(full_path):
                return full_path
            else:
                 print(f"Aviso: Método get_end não encontrou o arquivo diretamente. {full_path}")
                 # treatment = DataProcessing() # Talvez precise instanciar aqui?
                 raise FileNotFoundError(f"Arquivo não encontrado via get_end: {city_filename}")
        except Exception as e:
            self.controller.show_translated_message(
                msg_type='error',
                title_key='get_end_error_title',
                message_key='cant_get_path_msg',
                city_file_name=city_filename,
                error=str(e)
                )
            raise

    def _read_param(self, param_name, type_func, default_value=None):
        """Helper function to read and convert a parameter, with error handling."""
        try:
            # Assumes param_name matches the attribute name in current_param_frame
            widget_var = getattr(self.current_param_frame, param_name)
            value_str = widget_var.get()

            # Handle boolean separately if using Checkbuttons/BooleanVar
            if type_func == bool:
                return widget_var.get() # BooleanVar returns bool directly

            # Handle specific string values like 'auto', 'scale', etc.
            if type_func == str and value_str in ['auto', 'scale', 'sqrt', 'log2', 'none', 'best', 'random', 'mse', 'mae', 'linear', 'poly', 'rbf', 'sigmoid', 'lbfgs', 'sgd', 'adam', 'constant', 'invscaling', 'adaptive', 'identity', 'logistic', 'tanh', 'relu', 'ball_tree', 'kd_tree', 'brute']:
                 return value_str # Return known string values directly

            # Attempt conversion for numeric types
            if type_func == int:
                return int(value_str)
            if type_func == float:
                 # Try int first in case user enters "5" for a float field
                 try: return int(value_str)
                 except ValueError: return float(value_str)

            # If type_func is str but not a known keyword, return as is
            if type_func == str:
                 return value_str

            # Fallback for unexpected types
            return type_func(value_str) # General conversion attempt

        except AttributeError:
            self.controller.show_translated_message(
                msg_type='error',
                title_key='internal_error_title',
                message_key='atribute_not_found_msg',
                param_name=param_name
                )
            raise # Stop execution
        except (ValueError, TypeError) as e:
            self.controller.show_translated_message(
                msg_type='error',
                title_key='value_error_title',
                message_key='value_invalid_for_parameter_msg',
                param_name=param_name,
                type_func=type_func.__name__,
                value_str=value_str,
                error=str(e)
                )
            raise
        except Exception as e:
            self.controller.show_translated_message(
                msg_type='error',
                title_key='unexpected_error_title',
                message_key='error_while_reading_parameter_msg',
                param_name=param_name,
                error=str(e)
                )
            raise


    def generate_preview_dt(self):
        params = self._get_common_params()
        if params is None: return
        city_path, indicator, split_percentage, n_tests, save_model_flag = params

        try:
            criterion = self._read_param('criterion_v', str)
            splitter = self._read_param('splitter_v', str)
            max_depth = self._read_param('maxd_v', int)
            min_samples_split = self._read_param('minsam_s_v', int)
            min_samples_leaf = self._read_param('minsam_l_v', int)
            min_weight_fraction_leaf = self._read_param('minweifra_l_v', float)
            max_features = self.valid_maxf(self.current_param_frame.maxfeat_v.get())

            max_leaf_nodes = self._read_param('maxleaf_n', int)
            min_impurity_decrease = self._read_param('minimp_dec', float)
            ccp_alpha = self._read_param('ccp_alp_v', float)
        except Exception:
            return

        prev = Training()
        score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis = prev.decision_tree(
            city_path, indicator, split_percentage, criterion, splitter, max_depth,
            min_samples_leaf, max_features, max_leaf_nodes, n_tests, min_samples_split,
            min_weight_fraction_leaf, min_impurity_decrease, ccp_alpha, save_model_flag
        )
        self.data_preview(score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis)

    def generate_preview_bt(self):
        params = self._get_common_params()
        if params is None: return
        city_path, indicator, split_percentage, n_tests, save_model_flag = params

        try:
            # Parâmetros específicos do DT
            criterion = self._read_param('criterion_v', str)
            splitter = self._read_param('splitter_v', str)
            max_depth = self._read_param('maxd_v', int)
            min_samples_split = self._read_param('minsam_s_v', int)
            min_samples_leaf = self._read_param('minsam_l_v', int)
            min_weight_fraction_leaf = self._read_param('minweifra_l_v', float)
            max_features = self.valid_maxf(self.current_param_frame.maxfeat_v.get())
            max_leaf_nodes = self._read_param('maxleaf_n', int)
            min_impurity_decrease = self._read_param('minimp_dec', float)
            ccp_alpha = self._read_param('ccp_alp_v', float)
            # Parâmetro específico do Bagging
            n_estimators = self._read_param('n_estimators', int)
        except Exception:
            return

        prev = Training()
        score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis = prev.bagging_trees(
             city_path, indicator, split_percentage, criterion, splitter, max_depth,
             min_samples_leaf, max_features, max_leaf_nodes, n_tests, min_samples_split,
             min_weight_fraction_leaf, min_impurity_decrease, ccp_alpha, save_model_flag,
             n_estimators
        )
        self.data_preview(score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis)

    def generate_preview_nn(self):
        params = self._get_common_params()
        if params is None: return
        city_path, indicator, split_percentage, n_tests, save_model_flag = params

        try:
            activation = self._read_param('activation_v', str)
            solver = self._read_param('solver_v', str)
            alpha = self._read_param('alpha_v', float)
            batch_size_str = self._read_param('batch_size_v', str) # Read as string first
            learning_rate = self._read_param('learning_rate_v', str)
            learning_rate_init = self._read_param('learning_rate_init_v', float)
            power_t = self._read_param('power_t_v', float)
            max_iter = self._read_param('max_iter_v', int)
            shuffle = self._read_param('shuffle_v', bool) # Read BooleanVar
            tol = self._read_param('tol_v', float)
            verbose = self._read_param('verbose_v', bool) # Read BooleanVar
            warm_start = self._read_param('warm_start_v', bool) # Read BooleanVar
            momentum = self._read_param('momentum_v', float)
            nesterovs_momentum = self._read_param('nesterovs_momentum_v', bool) # Read BooleanVar
            early_stopping = self._read_param('early_stopping_v', bool) # Read BooleanVar
            validation_fraction = self._read_param('validation_fraction_v', float)
            beta_1 = self._read_param('beta_1_v', float)
            beta_2 = self._read_param('beta_2_v', float)
            n_iter_no_change = self._read_param('n_iter_no_change_v', int)
            max_fun = self._read_param('max_fun_v', int)

            # Converte batch_size se não for 'auto'
            batch_size = int(batch_size_str) if batch_size_str != 'auto' else 'auto'

        except Exception:
            return

        prev = Training()
        score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis = prev.neural_network(
             city_path, indicator, split_percentage, n_tests, activation, solver, alpha, batch_size,
             learning_rate, learning_rate_init, power_t, max_iter, shuffle, tol, verbose,
             warm_start, momentum, nesterovs_momentum, early_stopping, validation_fraction,
             beta_1, beta_2, n_iter_no_change, max_fun, save_model_flag
        )
        self.data_preview(score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis)

    def generate_preview_svm(self):
        params = self._get_common_params()
        if params is None: return
        city_path, indicator, split_percentage, n_tests, save_model_flag = params

        try:
            # ⚠️ Certifique-se que SVMParameterFrame foi refatorado e tem estes atributos
            kernel = self._read_param('kernel_v', str)
            degree = self._read_param('degree_v', int)
            gamma_str = self._read_param('gamma_v', str) # Read as string
            coef0 = self._read_param('coef0_v', float)
            tol = self._read_param('tol_v', float)
            c_param = self._read_param('c_v', float)
            epsilon = self._read_param('epsilon_v', float)
            shrinking = self._read_param('shrinking_v', bool) # BooleanVar?
            cache_size = self._read_param('cache_size_v', float)
            verbose = self._read_param('verbose_v', bool) # BooleanVar?
            max_iter = self._read_param('maxiter_v', int)

            # Converte gamma se necessário
            gamma = float(gamma_str) if gamma_str not in ['scale', 'auto'] else gamma_str

        except AttributeError:
             self.controller.show_translated_message(
                msg_type='error',
                title_key='internal_error_title',
                message_key='ui_problem_svm_msg',
                )
             return
        except Exception:
            return

        prev = Training()
        score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis = prev.support_vector_regression(
             city_path, indicator, split_percentage, n_tests, kernel, degree, gamma, coef0,
             tol, c_param, epsilon, shrinking, cache_size, verbose, max_iter, save_model_flag
        )
        self.data_preview(score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis)

    def generate_preview_GP(self):
        params = self._get_common_params()
        if params is None: return
        city_path, indicator, split_percentage, n_tests, save_model_flag = params

        try:
            
            alpha_gp = self._read_param('alpha_gp', float)
            n_restarts_optimizer = self._read_param('n_restarts_op', int)
            normalize_y_gp = self._read_param('normalize_y_gp', bool)
            kernel_type = self._read_param('kernel_type', str)

            length_scale = self._read_param('length_scale', float)
            
            nu = self._read_param('nu', float) 
            
            sigma_0 = self._read_param('sigma_0', float)
            alpha_rq = self._read_param('alpha_rq', float)

        except Exception:
            return

        prev = Training()
        
        # print(f"GP Call: Type={kernel_type}, Len={length_scale}, Alpha={alpha_gp}, Nu={nu}")

        try:
            score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis = prev.gaussian_process_regression(
                city=city_path, 
                indicator_code=indicator, 
                division=split_percentage, 
                kernel_type=kernel_type, 
                len_scale=length_scale,  
                nu=nu, 
                sigma_0=sigma_0, 
                alpha_rq=alpha_rq, 
                alpha_noise=alpha_gp,    
                n_restart=n_restarts_optimizer, 
                normalize_y=normalize_y_gp, 
                save_model=save_model_flag
            )
            
            self.data_preview(score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis)
        
        except TypeError as e:
            self.controller.show_translated_message(
                 msg_type='error',
                 title_key='execution_error_title',
                 message_key='model_error_msg',
                 option='Gaussian Process',
                 error=f"Erro de assinatura na função: {str(e)}"
             )
        except Exception as e:
            self.controller.show_translated_message(
                 msg_type='error',
                 title_key='execution_error_title',
                 message_key='model_error_msg',
                 option='Gaussian Process',
                 error=str(e)
             )
            
    def generate_preview_kn(self):
        params = self._get_common_params()
        if params is None: return
        city_path, indicator, split_percentage, n_tests, save_model_flag = params

        try:
            n_neighbors = self._read_param('n_neighbors_v', int)
            algorithm = self._read_param('algorithm_v', str)
            leaf_size = self._read_param('leaf_size_v', int)
            p_value = self._read_param('p_v', int)
            n_jobs_str = self._read_param('n_jobs_v', str)

            # Converte n_jobs
            n_jobs = None if n_jobs_str.lower() == 'none' else int(n_jobs_str)

        except Exception:
            return

        prev = Training()
        score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis = prev.KNeighbors(
             city_path, indicator, split_percentage, n_tests, n_neighbors, algorithm, leaf_size, p_value, n_jobs, save_model_flag
        )
        self.data_preview(score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis)


import json
import os

class LanguageManager:
    def __init__(self):
        self.translations = {}
        self.language_folder = os.path.dirname(os.path.abspath(__file__))
        try:
            self.available_languages = self._load_available_languages()
            self.current_language = None
        except FileNotFoundError:
            raise FileNotFoundError(
                f"A pasta de idiomas '{self.language_folder}' não foi encontrada."
            )

    def _load_available_languages(self):
        """Carrega os códigos de idioma disponíveis baseados nos nomes dos arquivos JSON."""
        langs = {}
        for filename in os.listdir(self.language_folder):
            if filename.endswith('.json'):
                lang_code = filename.split('.')[0]
                lang_map = {"pt_br": "Português (BR)", "en_us": "English (US)", "es_es": "Español (ES)"}
                langs[lang_map.get(lang_code, lang_code)] = lang_code
        return langs
    
    def set_language(self, lang_code):
        """Define o idioma atual e carrega suas traduções"""
        file_path = os.path.join(self.language_folder, f"{lang_code}.json")
        try:
            with open(file_path, 'r', encoding='utf-8') as f:
                self.translations = json.load(f)
                self.current_language = lang_code
        except FileNotFoundError:
            print(f"Arquivo de idioma '{lang_code}.json' não encontrado.")
            if self.current_language is None:
                self.set_language("pt_br")

    def get(self, key, **kwargs):
        """Retorna o texto traduzido para uma chave específica"""
        text = self.translations.get(key, key)
        if kwargs:
            try:
                return text.format(**kwargs)
            except KeyError:
                return text 
        return text


def get_parameters_ml(self):
    return {
        "cidade": self.get_end(self.data_s.get()),
        "save_model": self.save_model.get(),
        "num_testes": int(self.num_teste.get()),
        "porcentagem_treinamento": int(self.por_trei.get()),
        "n_neighbors": self.n_neighbors_v.get(),
        "algorithm": self.algorithm_v.get(),
        "leaf_size": self.leaf_size_v.get(),
        "p_v": self.p_v.get(),
        "n_jobs": int(self.n_jobs_v.get()) if self.n_jobs_v.get().isdigit() else self.n_jobs_v.get()
    }


import pandas as pd
import numpy as np
from datetime import datetime, timedelta

def make_database(data, file_name):
    start_date = pd.to_datetime(data[0][0])
    end_date = pd.to_datetime(data[-1][0])
    
    date_range = pd.date_range(start=start_date, end=end_date, freq='D')
    
    data_base = []
    data_index = 0

    for date in date_range:
        found = False
        
        while data_index < len(data):
            record_date = pd.to_datetime(data[data_index][0])
            
            if record_date == date:
                data_base.append([
                    date.year, date.month, date.day,
                    data[data_index][1],
                    data[data_index][2],
                    data[data_index][3]
                ])
                data_index += 1
                found = True
                break
            
            data_index += 1

        if not found:
            data_base.append([date.year, date.month, date.day, -99, -99, -99])
    
    df = pd.DataFrame(data_base, columns=['Year', 'Month', 'Day', 'V1', 'V2', 'V3'])
    df.to_excel(file_name, index=False)

    return df


def clean_missing_data(df):
    return df[(df[['V', 'V2', 'V3']] != -99).all(axis=1)]


def normalize_data(df):
    normalized_df = df.copy()
    
    for col in ['V1', 'V2', 'V3']:
        if df[col].min() != df[col].max():
            normalized_df[col] = ((df[col] - df[col].min()) / (df[col].max() - df[col].min())) * 0.6 + 0.2
    
    return normalized_df




import time
import pickle
from sklearn.tree import DecisionTreeRegressor
from typing import Tuple, List
import os
import numpy as np
from .utils import calculate_errors

def decision_tree_tr(
    self, city: str, indicator_code: int, split_ratio: float, criterion: str, splitter: str,
    max_depth: int, min_samples_leaf: int, max_features, max_leaf_nodes: int,
    n_tests: int, min_samples_split: int, min_weight_fraction_leaf: float,
    min_impurity_decrease: float, ccp_alpha: float, save_model: bool
) -> Tuple[float, float, float, float, float, float, float, float, float, List[float], List[float], List[int]]:
    """
    Trains and evaluates a Decision Tree Regressor multiple times and returns statistics.

    Returns:
        Tuple containing: 
            score, average absolute error, average relative error,
            max absolute error, exact value at max error, predicted value at max error,
            min absolute error, exact value at min error, predicted value at min error,
            list of exact values, list of predicted values, x axis indices
    """
    start_time = time.time()

    indicators = {3: 'Precipitation', 4: 'Maximum Temperature'}
    indicator = indicators.get(indicator_code, 'Minimum Temperature')

    train_X, train_y, val_X, val_y = self.prepare_matrix_by_city(city, split_ratio, indicator_code)

    all_exact = []
    all_predicted = []
    total_relative_error = 0
    total_absolute_error = 0

    for test in range(n_tests):
        model = DecisionTreeRegressor(
            criterion=criterion, splitter=splitter, max_depth=max_depth,
            min_samples_leaf=min_samples_leaf, max_features=max_features,
            max_leaf_nodes=max_leaf_nodes, min_samples_split=min_samples_split,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            min_impurity_decrease=min_impurity_decrease, ccp_alpha=ccp_alpha
        )
        model.fit(train_X, train_y)

        predictions = model.predict(val_X)
        absolute_errors = np.abs(np.array(val_y) - predictions)
        relative_errors = absolute_errors / np.array(val_y)

        total_absolute_error += np.mean(absolute_errors)
        total_relative_error += np.mean(relative_errors)

        if test != n_tests:  # This check is always True, unless intended to avoid the last?
            all_exact.extend(val_y)
            all_predicted.extend(predictions)

    statistics = calculate_errors(all_exact, all_predicted, total_absolute_error, total_relative_error, n_tests)

    score = statistics["score"]
    max_error = statistics["max_error"]
    min_error = statistics["min_error"]
    exact_max = statistics["exact_max"]
    predicted_max = statistics["predicted_max"]
    exact_min = statistics["exact_min"]
    predicted_min = statistics["predicted_min"]
    avg_absolute_error = statistics["avg_absolute_error"]
    avg_relative_error = statistics["avg_relative_error"]
    
    if save_model:
        with open(os.path.join(os.getcwd(), 'modelo_rn.sav'), 'wb') as f:
                pickle.dump(model, f)

    x_axis = list(range(1, len(all_exact) + 1))

    return (
        score, avg_absolute_error, avg_relative_error,
        max_error, exact_max, predicted_max,
        min_error, exact_min, predicted_min,
        all_exact, all_predicted, x_axis
    )
